# -*- coding: utf-8 -*-
"""
users internal manager module.
"""

import pyrin.security.services as security_services
import pyrin.validator.services as validator_services
import pyrin.security.hashing.services as hashing_services
import pyrin.globalization.datetime.services as datetime_services

from pyrin.core.globals import _
from pyrin.core.structs import Manager
from pyrin.users.internal import InternalUsersPackage
from pyrin.users.internal.models import InternalUserEntity
from pyrin.database.services import get_current_store
from pyrin.users.internal.exceptions import InternalUserNotFoundError, PasswordsDoNotMatchError


class InternalUsersManager(Manager):
    """
    internal users manager class.
    """

    package_class = InternalUsersPackage

    def _get(self, id, **options):
        """
        gets the specified internal user.

        :param int id: internal user id to get its info.

        :raises InternalUserNotFoundError: internal user not found error.

        :rtype: InternalUserEntity
        """

        data = validator_services.validate(InternalUserEntity, id=id)
        store = get_current_store()
        user = store.query(InternalUserEntity).get(data.id)

        if user is None:
            raise InternalUserNotFoundError(_('Internal user [{user_id}] not found.')
                                            .format(user_id=data.id))

        return user

    def _validate_passwords(self, password, confirm_password):
        """
        validates that given passwords are match.

        :param str password: password.
        :param str confirm_password: confirm password.

        :raises PasswordsDoNotMatchError: passwords do not match error.
        """

        if password != confirm_password:
            message = _('Provided passwords do not match')
            raise PasswordsDoNotMatchError(f'{message}.',
                                           data=dict(password=message,
                                                     confirm_password=message))

    def get(self, id, *columns, **options):
        """
        gets the internal user with given id.

        :param int id: internal user id to be get.

        :param CoreColumn columns: columns to be fetched.
                                   if not provided all columns will be fetched.

        :raises InternalUserNotFoundError: internal user not found error.

        :rtype: InternalUserEntity | ROW_RESULT
        """

        if not columns:
            return self._get(id, **options)

        data = validator_services.validate(InternalUserEntity, id=id)
        store = get_current_store()
        user = store.query(*columns).filter(InternalUserEntity.id == data.id).one_or_none()
        if user is None:
            raise InternalUserNotFoundError(_('Internal user [{user_id}] not found.')
                                            .format(user_id=data.id))

        return user

    def is_active(self, id, **options):
        """
        gets a value indicating that given internal user is active.

        :param int id: internal user id to check its active status.

        :raises InternalUserNotFoundError: internal user not found error.

        :rtype: bool
        """

        entity = self.get(id, InternalUserEntity.is_active, **options)
        return entity.is_active

    def create(self, username, password, confirm_password,
               first_name, last_name, **options):
        """
        creates a new internal user based on given inputs.

        :param str username: username.
        :param str password: password.
        :param str confirm_password: confirm password.
        :param str first_name: first name.
        :param str last_name: last name.

        :keyword str mobile: mobile number.
        :keyword str email: email address.
        :keyword int gender: gender.
        :enum gender:
            FEMALE = 0
            MALE = 1
            OTHER = 2

        :keyword bool is_active: is active user.
        :keyword bool is_superuser: is superuser.

        :raises PasswordsDoNotMatchError: passwords do not match error.
        """

        options.update(username=username, password=password,
                       confirm_password=confirm_password,
                       first_name=first_name, last_name=last_name)
        validator_services.validate_dict(InternalUserEntity, options)
        password = options.get('password')
        confirm_password = options.get('confirm_password')
        self._validate_passwords(password, confirm_password)
        entity = InternalUserEntity(**options)
        entity.password_hash = security_services.get_password_hash(password)
        entity.save()

    def update(self, id, **options):
        """
        updates the given internal user based on given inputs.

        :param int id: internal user id.

        :keyword str username: username.
        :keyword str password: password.
        :keyword str confirm_password: confirm password.
        :keyword str first_name: first name.
        :keyword str last_name: last name.
        :keyword str mobile: mobile number.
        :keyword str email: email address.
        :keyword int gender: gender.
        :enum gender:
            FEMALE = 0
            MALE = 1
            OTHER = 2

        :keyword bool is_active: is active user.
        :keyword bool is_superuser: is superuser.

        :raises PasswordsDoNotMatchError: passwords do not match error.
        """

        entity = self._get(id)
        validator_services.validate_dict(InternalUserEntity, options, for_update=True)
        password = options.get('password')
        confirm_password = options.get('confirm_password')
        if password is not None or confirm_password is not None:
            self._validate_passwords(password, confirm_password)
            entity.password_hash = security_services.get_password_hash(password)

        entity.update(**options)

    def get_login_user(self, username, password, *columns, **options):
        """
        gets an internal user with given username and password for logging in.

        it may return None if no internal user found.

        :param str username: username.
        :param str password: password.

        :param CoreColumn columns: columns to be fetched.
                                   if not provided all columns will be fetched.

        :rtype: InternalUserEntity | ROW_RESULT
        """

        columns = list(columns)
        if not columns:
            columns = [InternalUserEntity]
        elif InternalUserEntity.password_hash not in columns:
            columns.append(InternalUserEntity.password_hash)

        store = get_current_store()
        user = store.query(*columns)\
            .filter(InternalUserEntity.username == username).one_or_none()

        if user is None or hashing_services.is_match(password, user.password_hash) is not True:
            return None

        return user

    def update_last_login_at(self, id, **options):
        """
        updates the last login at for given user to current datetime.

        :param int id: internal user id to update its last login at.

        :raises InternalUserNotFoundError: internal user not found error.
        """

        store = get_current_store()
        count = store.query(InternalUserEntity).filter(InternalUserEntity.id == id)\
            .update({InternalUserEntity.last_login_at: datetime_services.now()})

        if count <= 0:
            raise InternalUserNotFoundError(_('Internal user [{user_id}] not found.')
                                            .format(user_id=id))



# -*- coding: utf-8 -*-
"""
audit security authenticators module.
"""

from pyrin.security.enumerations import InternalAuthenticatorEnum
from pyrin.security.authentication.decorators import authenticator
from pyrin.security.authentication.handlers.internal import InternalTokenAuthenticator


@authenticator()
class AuditTokenAuthenticator(InternalTokenAuthenticator):
    """
    audit token authenticator class.
    """

    REFRESH_TOKEN_HOLDER = 'Audit-Refresh-Auth'
    _name = InternalAuthenticatorEnum.AUDIT

    def _get_extra_user_info(self, user, **options):
        """
        gets extra user info for given internal user if required.

        :param ROW_RESULT user: internal user to get its extra info.

        :rtype: dict
        """

        return dict(audit_access=user.audit_access)



# -*- coding: utf-8 -*-
"""
swagger tags exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SwaggerTagsException(CoreException):
    """
    swagger tags exception.
    """
    pass


class TagNameIsRequiredError(SwaggerTagsException):
    """
    tag name is required error.
    """
    pass


class TagIsRequiredError(SwaggerTagsException):
    """
    tag is required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
admin api module.
"""

import pyrin.admin.services as admin_services

from pyrin.api.router.decorators import api, post, patch, delete


admin_config = admin_services.get_admin_configurations()
url = admin_services.get_admin_base_url()
admin_config.update(swagger=False)
admin_config.pop('paged', None)
admin_config.pop('readable', None)
admin_config.pop('url', None)
is_enabled = admin_config.pop('enabled', False)

if is_enabled is True:
    @api(f'{url}<register_name>/<pk>', **admin_config)
    def get(register_name, pk, **options):
        """
        gets an entity with given primary key.

        :param str register_name: register name of admin page.
        :param object pk: primary key of entity to be get.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        :raises EntityNotFoundError: entity not found error.

        :rtype: pyrin.database.model.base.BaseEntity
        """

        return admin_services.get(register_name, pk)


    @api(f'{url}<register_name>', **admin_config)
    def find(register_name, **filters):
        """
        performs find on given admin page and returns the result.

        :param str register_name: register name of admin page.

        :keyword **filters: all filters to be passed to related admin page.

        :rtype: list[ROW_RESULT]
        """

        return admin_services.find(register_name, **filters)


    @post(f'{url}<register_name>', **admin_config)
    def create(register_name, **data):
        """
        performs create on given admin page.

        :param str register_name: register name of admin page.

        :keyword **data: all data to be passed to related admin page for data creation.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.

        :rtype: object
        """

        return admin_services.create(register_name, **data)


    @patch(f'{url}<register_name>/<pk>', **admin_config)
    def update(register_name, pk, **data):
        """
        performs update on given admin page.

        :param str register_name: register name of admin page.
        :param object pk: entity primary key to be updated.

        :keyword **data: all data to be passed to related admin page for data creation.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        :raises EntityNotFoundError: entity not found error.
        """

        return admin_services.update(register_name, pk, **data)


    @delete(f'{url}<register_name>/<pk>', **admin_config)
    def remove(register_name, pk, **options):
        """
        performs remove on given admin page.

        :param str register_name: register name of admin page.
        :param object pk: entity primary key to be removed.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        return admin_services.remove(register_name, pk)


    @delete(f'{url}<register_name>/bulk', **admin_config)
    def remove_bulk(register_name, pk):
        """
        performs remove bulk on given admin page.

        :param str register_name: register name of admin page.
        :param object | list[object] pk: entity primary keys to be removed.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        return admin_services.remove_bulk(register_name, pk)


    @delete(f'{url}<register_name>', **admin_config)
    def remove_all(register_name):
        """
        performs remove all on given admin page.

        :param str register_name: register name of admin page.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        return admin_services.remove_all(register_name)


    @api(f'{url}metadata', **admin_config)
    def get_main_metadata(**options):
        """
        gets all admin pages main metadata.

        :raises AdminPagesHaveNotLoadedError: admin pages have not loaded error.

        :returns: dict(list pages,
                       dict configs)
        :rtype: dict
        """

        return admin_services.get_main_metadata()


    @api(f'{url}metadata/<register_name>/find', **admin_config)
    def get_find_metadata(register_name, **options):
        """
        gets the find metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        return admin_services.get_find_metadata(register_name)


    @api(f'{url}metadata/<register_name>/create', **admin_config)
    def get_create_metadata(register_name, **options):
        """
        gets the create metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        return admin_services.get_create_metadata(register_name)


    @api(f'{url}metadata/<register_name>/update', **admin_config)
    def get_update_metadata(register_name, **options):
        """
        gets the update metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        return admin_services.get_update_metadata(register_name)


    @api(f'{url}metadata/configs', **admin_config)
    def get_configs(**options):
        """
        gets the required configs of admin api.

        :returns: dict(str panel_name,
                       str page_key,
                       str page_size_key,
                       str ordering_key,
                       str locale_key,
                       str timezone_key,
                       str search_param,
                       str hidden_pk_name)
        :rtype: dict
        """

        return admin_services.get_configs()

    @post(f'{url}login', authenticated=False)
    def login(username, password, **options):
        """
        logs in an internal user with given info into admin panel.

        :param str username: username.
        :param str password: password.

        :raises ProvidedUsernameOrPasswordAreIncorrect: provided username or
                                                        password are incorrect.

        :returns: dict(str access_token)
        :rtype: dict
        """

        return admin_services.login(username, password, **options)



# -*- coding: utf-8 -*-
"""
authentication services module.
"""

from pyrin.application.services import get_component
from pyrin.security.authentication import AuthenticationPackage


def get_relevant_authenticator_name(url):
    """
    gets the relevant authenticator name for given url rule.

    it looks for relevant authenticator in the following order:

    1. looking for it in rule based authenticators.
    2. getting default authenticator for it.

    it may return None if no relevant authenticator could be found.

    :param str url: url rule to find a relevant authenticator for it.

    :rtype: str
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME)\
        .get_relevant_authenticator_name(url)


def get_current_authenticator_name():
    """
    gets the authenticator name for current request from its matched url rule.

    it may return None if the current request does not match any url rule.
    it also returns None for routes which have `authenticated=False` in their
    definition.

    :rtype: str
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).get_current_authenticator_name()


def get_current_authenticator():
    """
    gets the authenticator for current request from its matched url rule.

    it may return None if the current request does not match any url rule.
    it also returns None for routes which have `authenticated=False` in their
    definition.

    :rtype: AbstractAuthenticatorBase
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).get_current_authenticator()


def get_rule_based_authenticator_name(url):
    """
    gets the relevant authenticator name to the given url rule.

    it may return None if no matching rule is found.

    :param str url: url to get its matching authenticator name.

    :rtype: str
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME)\
        .get_rule_based_authenticator_name(url)


def authenticator_exists(name):
    """
    gets a value indicating that an authenticator with given name exists.

    :param str name: authenticator name.

    :rtype: bool
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).authenticator_exists(name)


def get_authenticator(name, **options):
    """
    gets the authenticator with given name.

    :keyword str name: authenticator name.

    :raises AuthenticatorNotFoundError: authenticator not found error.

    :rtype: AbstractAuthenticatorBase
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).get_authenticator(name, **options)


def register_authenticator(instance, **options):
    """
    registers a new authenticator or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an authenticator which is already registered.

    :param AbstractAuthenticatorBase instance: authenticator to be registered.
                                               it must be an instance of
                                               AbstractAuthenticatorBase.

    :keyword bool replace: specifies that if there is another registered
                           authenticator with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidAuthenticatorTypeError: invalid authenticator type error.
    :raises DuplicatedAuthenticatorError: duplicated authenticator error.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).register_authenticator(instance,
                                                                                      **options)


def validate_authenticators():
    """
    validates that all protected routes authenticators are present.

    :raises AuthenticatorNotFoundError: authenticator not found error.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).validate_authenticators()


def authenticate(request, **options):
    """
    authenticates given request and pushes the authenticated data into current request.

    if authentication fails, authenticated data will not be pushed into current request.

    :param CoreRequest request: request to be authenticated.

    :keyword str authenticator: authenticator name to be used. if not
                                provided, current request authenticator
                                will be used. if current request does
                                not have an authenticator, nothing will
                                be done.

    :raises AuthenticationFailedError: authentication failed error.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).authenticate(request,
                                                                            **options)


def login(username, password, authenticator, **options):
    """
    logs in a user with given info using provided authenticator.

    it may return the required credentials if they must be returned to client.

    :param str username: username.
    :param str password: password.
    :param str authenticator: authenticator name to be used.

    :raises ValidationError: validation error.
    :raises AuthenticatorNotFoundError: authenticator not found error.

    :returns: required credentials.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).login(username, password,
                                                                     authenticator, **options)


def logout(authenticator, **options):
    """
    logouts the current user and clears its relevant credentials.

    :param str authenticator: authenticator name to be used.

    :raises AuthenticatorNotFoundError: authenticator not found error.
    """

    return get_component(AuthenticationPackage.COMPONENT_NAME).logout(authenticator, **options)



# -*- coding: utf-8 -*-
"""
users internal services module.
"""

from pyrin.application.services import get_component
from pyrin.users.internal import InternalUsersPackage


def get(id, *columns, **options):
    """
    gets the internal user with given id.

    :param int id: internal user id to be get.

    :param CoreColumn columns: columns to be fetched.
                               if not provided all columns will be fetched.

    :raises InternalUserNotFoundError: internal user not found error.

    :rtype: InternalUserEntity | ROW_RESULT
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).get(id, *columns, **options)


def is_active(id, **options):
    """
    gets a value indicating that given internal user is active.

    :param int id: internal user id to check its active status.

    :raises InternalUserNotFoundError: internal user not found error.

    :rtype: bool
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).is_active(id, **options)


def create(username, password, confirm_password, first_name, last_name, **options):
    """
    creates a new internal user based on given inputs.

    :param str username: username.
    :param str password: password.
    :param str confirm_password: confirm password.
    :param str first_name: first name.
    :param str last_name: last name.

    :keyword str mobile: mobile number.
    :keyword str email: email address.
    :keyword int gender: gender.
    :enum gender:
        FEMALE = 0
        MALE = 1
        OTHER = 2

    :keyword bool is_active: is active user.
    :keyword bool is_superuser: is superuser.

    :raises PasswordsDoNotMatchError: passwords do not match error.
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).create(username, password,
                                                                     confirm_password,
                                                                     first_name, last_name,
                                                                     **options)


def update(id, **options):
    """
    updates the given internal user based on given inputs.

    :param int id: internal user id.

    :keyword str username: username.
    :keyword str password: password.
    :keyword str confirm_password: confirm password.
    :keyword str first_name: first name.
    :keyword str last_name: last name.
    :keyword str mobile: mobile number.
    :keyword str email: email address.
    :keyword int gender: gender.
    :enum gender:
        FEMALE = 0
        MALE = 1
        OTHER = 2

    :keyword bool is_active: is active user.
    :keyword bool is_superuser: is superuser.

    :raises PasswordsDoNotMatchError: passwords do not match error.
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).update(id, **options)


def get_login_user(username, password, *columns, **options):
    """
    gets an internal user with given username and password for logging in.

    it may return None if no internal user found.

    :param str username: username.
    :param str password: password.

    :param CoreColumn columns: columns to be fetched.
                               if not provided all columns will be fetched.

    :rtype: InternalUserEntity | ROW_RESULT
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).get_login_user(username, password,
                                                                             *columns, **options)


def update_last_login_at(id, **options):
    """
    updates the last login at for given user to current datetime.

    :param int id: internal user id to update its last login at.

    :raises InternalUserNotFoundError: internal user not found error.
    """

    return get_component(InternalUsersPackage.COMPONENT_NAME).update_last_login_at(id, **options)



# -*- coding: utf-8 -*-
"""
authentication models module.
"""

from sqlalchemy import BigInteger

from pyrin.database.enumerations import FKRestrictionEnum
from pyrin.database.model.mixin import HistoryMixin
from pyrin.database.model.declarative import CoreEntity
from pyrin.security.enumerations import TokenTypeEnum, InternalAuthenticatorEnum
from pyrin.database.orm.sql.schema.columns import GUIDColumn, StringColumn, \
    FKColumn, BooleanColumn


class InternalSessionBaseEntity(CoreEntity):
    """
    internal session base entity class.
    """

    _table = 'internal_session'

    id = GUIDColumn(name='id', primary_key=True, index=True)


class InternalSessionEntity(InternalSessionBaseEntity, HistoryMixin):
    """
    internal session entity class.
    """

    _extend_existing = True

    user_id = FKColumn(name='user_id', type_=BigInteger, fk='internal_user.id',
                       fk_on_delete=FKRestrictionEnum.CASCADE)
    type = StringColumn(name='type', check_in=TokenTypeEnum, nullable=False)
    authenticator = StringColumn(name='authenticator', nullable=False,
                                 check_in=InternalAuthenticatorEnum)
    is_revoked = BooleanColumn(name='is_revoked', nullable=False, default=False)



# -*- coding: utf-8 -*-
"""
authorization manager module.
"""

import pyrin.application.services as application_services
import pyrin.security.session.services as session_services
import pyrin.security.authentication.services as authentication_services

from pyrin.core.globals import _
from pyrin.core.structs import Manager, Context
from pyrin.utils.custom_print import print_warning
from pyrin.security.authorization import AuthorizationPackage
from pyrin.api.router.handlers.protected import ProtectedRoute
from pyrin.security.authorization.interface import AbstractAuthorizerBase
from pyrin.security.exceptions import AuthorizationFailedError
from pyrin.security.authorization.exceptions import UserNotAuthenticatedError, \
    AuthorizerNotFoundError, InvalidAuthorizerTypeError, DuplicatedAuthorizerError


class AuthorizationManager(Manager):
    """
    authorization manager class.
    """

    package_class = AuthorizationPackage

    def __init__(self):
        """
        initializes an instance of AuthorizationManager.
        """

        super().__init__()

        # a dictionary containing information of registered authorizers.
        # example: dict(str name: AbstractAuthorizerBase instance)
        self._authorizers = Context()

    def authorize(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will consider it as authorized.

        :keyword dict user_info: user info to be used for authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        if user is None:
            raise UserNotAuthenticatedError(_('User has not been authenticated.'))

        authorizer_name = options.pop('authorizer', None)
        if authorizer_name is None:
            authorizer_name = authentication_services.get_current_authenticator_name()

        if authorizer_name is None or not self.authorizer_exists(authorizer_name):
            return

        authorizer = self.get_authorizer(authorizer_name)
        try:
            authorizer.authorize(user, permissions, **options)
        except AuthorizationFailedError:
            raise
        except Exception as error:
            raise AuthorizationFailedError(error) from error

    def is_authorized(self, permissions, user=None, **options):
        """
        gets a value indicating that specified user is authorized for given permissions.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for authorization.

        :param user: user identity to be checked for authorization.
                     if not provided, current user will be used.

        :keyword str authorizer: authorizer name to be used.
                                 if not provided, current request authenticator will
                                 be used. if current request does not have an
                                 authenticator, it will return True as authorized.

        :keyword dict user_info: user info to be used for authorization.
                                 if not provided, current user info will be used.

        :rtype: bool
        """

        if user is None:
            user = session_services.get_current_user()

        user_info = options.get('user_info')
        if user_info is None:
            user_info = session_services.get_current_user_info()

        try:
            options.update(user_info=user_info)
            self.authorize(user, permissions, **options)
            return True
        except AuthorizationFailedError:
            return False

    def authorizer_exists(self, name):
        """
        gets a value indicating that an authorizer with given name exists.

        :param str name: authorizer name.

        :rtype: bool
        """

        return name in self._authorizers

    def get_authorizer(self, name, **options):
        """
        gets the authorizer with given name.

        :keyword str name: authorizer name.

        :raises AuthorizerNotFoundError: authorizer not found error.

        :rtype: AbstractAuthorizerBase
        """

        if name not in self._authorizers:
            raise AuthorizerNotFoundError('Authorizer [{name}] not found.'
                                          .format(name=name))

        return self._authorizers[name]

    def register_authorizer(self, instance, **options):
        """
        registers a new authorizer or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an authorizer which is already registered.

        :param AbstractAuthorizerBase instance: authorizer to be registered.
                                                it must be an instance of
                                                AbstractAuthorizerBase.

        :keyword bool replace: specifies that if there is another registered
                               authorizer with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidAuthorizerTypeError: invalid authorizer type error.
        :raises DuplicatedAuthorizerError: duplicated authorizer error.
        """

        if not isinstance(instance, AbstractAuthorizerBase):
            raise InvalidAuthorizerTypeError('Input parameter [{instance}] is '
                                             'not an instance of [{base}].'
                                             .format(instance=instance,
                                                     base=AbstractAuthorizerBase))

        if instance.name in self._authorizers:
            old_instance = self._authorizers.get(instance.name)
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedAuthorizerError('There is another registered '
                                                'authorizer [{old}] with name '
                                                '[{name}] but "replace" option is not '
                                                'set, so authorizer [{instance}] '
                                                'could not be registered.'
                                                .format(old=old_instance,
                                                        name=instance.name,
                                                        instance=instance))

            print_warning('Authorizer [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._authorizers[instance.name] = instance

    def validate_authorizers(self):
        """
        validates that all routes with permissions have their authorizers present.

        :raises AuthorizerNotFoundError: authorizer not found error.
        """

        app = application_services.get_current_app()
        for route in app.url_map.get_routes():
            if isinstance(route, ProtectedRoute) and route.permissions and \
                    not self.authorizer_exists(route.authenticator):
                raise AuthorizerNotFoundError('Authorizer with name [{name}] on '
                                              'view function [{endpoint}] does not '
                                              'exist in registered authorizers.'
                                              .format(name=route.authenticator,
                                                      endpoint=route.endpoint))



# -*- coding: utf-8 -*-
"""
admin interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class AdminPageSingletonMeta(MultiSingletonMeta):
    """
    admin page singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractAdminPage(CoreObject, metaclass=AdminPageSingletonMeta):
    """
    abstract admin page class.
    """

    @abstractmethod
    def get_entity(self):
        """
        gets the entity class of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type[pyrin.database.model.base.BaseEntity]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_register_name(self):
        """
        gets the register name of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_category(self):
        """
        gets the category of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_plural_name(self):
        """
        gets the plural name of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_singular_name(self):
        """
        gets the singular name of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get(self, pk):
        """
        gets an entity with given primary key.

        :param object pk: primary key of entity to be get.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: pyrin.database.model.base.BaseEntity
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def find(self, **filters):
        """
        finds entities with given filters.

        :keyword **filters: all filters to be passed to related find service.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: list[ROW_RESULT]
        """

        raise CoreNotImplementedError()

    @classmethod
    @abstractmethod
    def create(cls, **data):
        """
        creates an entity with given data.

        it's preferred for this method to return the pk of created entity
        if it is not a composite primary key. this lets the client to fill
        fk fields automatically after create.

        :keyword **data: all data to be passed to related create service.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @classmethod
    @abstractmethod
    def update(cls, pk, **data):
        """
        updates an entity with given data.

        :param object pk: entity primary key to be updated.

        :keyword **data: all data to be passed to related update service.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @classmethod
    @abstractmethod
    def remove(cls, pk):
        """
        deletes an entity with given pk.

        :param object pk: entity primary key to be deleted.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def remove_bulk(self, pk):
        """
        deletes entities with given primary keys.

        :param object | list[object] pk: entity primary keys to be deleted.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def remove_all(self):
        """
        deletes all entities.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def call_method(self, name, argument):
        """
        calls the method with given name with given argument and returns the result.

        :param str name: method name.
        :param ROW_RESULT argument: the method argument.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_get_permission(self):
        """
        gets a value indicating that this admin page has get permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_create_permission(self):
        """
        gets a value indicating that this admin page has create permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_update_permission(self):
        """
        gets a value indicating that this admin page has update permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_remove_permission(self):
        """
        gets a value indicating that this admin page has single or bulk remove permission.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def has_remove_all_permission(self):
        """
        gets a value indicating that this admin page has remove all permission.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_main_metadata(self):
        """
        gets the main metadata of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_find_metadata(self):
        """
        gets the find metadata of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_create_metadata(self):
        """
        gets the create metadata of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_update_metadata(self):
        """
        gets the update metadata of this admin page.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def populate_caches(self):
        """
        populates required caches of this admin page.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def method_names(self):
        """
        gets the list of all method names of this admin page to be used for result processing.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: tuple[str]
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
validator interface module.
"""

from abc import abstractmethod

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.structs import CoreObject


class AbstractValidatorBase(CoreObject):
    """
    abstract validator base class.

    all application validators must be subclassed from this.
    """

    @abstractmethod
    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object | list[object] value: value to be validated.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :keyword bool for_find: specifies that validation is for find operation.
                                defaults to False if not provided.
                                if this validator is for find and `for_find=False`
                                is provided, no validation will be done.

        :keyword str field_name: a custom field name to be used in validation errors.

        :keyword bool ignore_default: do not consider default value of the column
                                      for nullability validation. defaults to False
                                      if not provided and if the value is None but
                                      the column has default value it will be
                                      considered as valid.

        :raises CoreNotImplementedError: core not implemented error.
        :raises ValidationError: validation error.

        :returns: object | list[object]
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def get_info(self, for_update=False):
        """
        gets the info of this validator.

        :param bool for_update: specifies that info must be for update operation.
                                defaults to False if not provided and the info of
                                create will be returned.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: dict(bool required: is required for specified operation,
                       str form_field_type: form field type)

        :rtype: dict
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this validator.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.
        if a validator is not specific to an entity, then
        domain could be a unique string name.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: type[BaseEntity] | str
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def for_find(self):
        """
        gets a value indicating that this validator should only be used on validation for find.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
orm sql schema globals module.
"""

SMALL_INTEGER_MAX = 32767
SMALL_INTEGER_MIN = -32768
INTEGER_MAX = 2147483647
INTEGER_MIN = -2147483648
BIG_INTEGER_MAX = 9223372036854775807
BIG_INTEGER_MIN = -9223372036854775808



# -*- coding: utf-8 -*-
"""
admin security authorizers module.
"""

from pyrin.core.globals import _
from pyrin.security.authorization.decorators import authorizer
from pyrin.security.enumerations import InternalAuthenticatorEnum
from pyrin.security.authorization.handlers.internal import InternalAuthorizer
from pyrin.admin.security.exceptions import AdminAccessNotAllowedError


@authorizer()
class AdminAuthorizer(InternalAuthorizer):
    """
    admin authorizer class.
    """

    _name = InternalAuthenticatorEnum.ADMIN

    def _authorize_access(self, user, **options):
        """
        authorizes the given internal user for admin access.

        :param int user: internal user id to be checked if it is authorized.

        :keyword dict user_info: internal user info to be used to check for user.

        :raises AdminAccessNotAllowedError: admin access not allowed error.
        """

        user_info = options.get('user_info')
        if not user_info or user_info.get('admin_access') is not True:
            raise AdminAccessNotAllowedError(_('You are not allowed to access the admin panel. '
                                               'If you think that this is a mistake, please '
                                               'contact the support team.'))



# -*- coding: utf-8 -*-
"""
admin manager module.
"""

from copy import deepcopy

import pyrin.utils.dictionary as dict_utils
import pyrin.configuration.services as config_services
import pyrin.database.paging.services as paging_services
import pyrin.database.services as database_services
import pyrin.globalization.datetime.services as datetime_services
import pyrin.globalization.locale.services as locale_services
import pyrin.security.authentication.services as authentication_services

from pyrin.core.globals import _
from pyrin.admin import AdminPackage
from pyrin.core.structs import Context, Manager
from pyrin.admin.interface import AbstractAdminPage
from pyrin.security.enumerations import InternalAuthenticatorEnum
from pyrin.admin.enumerations import ListFieldTypeEnum, FormFieldTypeEnum
from pyrin.admin.exceptions import InvalidAdminPageTypeError, DuplicatedAdminPageError, \
    AdminPageNotFoundError, AdminOperationNotAllowedError, AdminPagesHaveNotLoadedError


class AdminManager(Manager):
    """
    admin manager class.
    """

    package_class = AdminPackage

    def __init__(self, **options):
        """
        initializes an instance of AdminManager.
        """

        super().__init__()

        # a dict containing all registered admin pages in the form of:
        # {str register_name: AbstractAdminPage instance}
        self._admin_pages = Context()

        # a dict containing all registered admin pages for different entity types.
        # in the form of:
        # {BaseEntity entity: AbstractAdminPage instance}
        self._admin_entities = Context()

        # a tuple of all available admin pages metadata sorted by category and name.
        # in the form of:
        # ({str category: [dict admin_metadata]})
        self._admin_metadata = None

        # a dict containing a map between all form field types and list field types.
        # for example: {str form_field_type: str list_field_type}
        self._type_map = self._get_form_to_list_type_map()

        self._base_url = self._load_base_url()

        # shared admin panel configs required for client.
        self._configs = None

    def _get_form_to_list_type_map(self):
        """
        gets the type map for different form fields to list fields.

        **NOTE:**

        as the client side table does not format the numeric values correctly, we
        have to introduce numeric values as string to the client to keep the behavior
        of these column types as others.

        :rtype: dict
        """

        result = Context()
        result[FormFieldTypeEnum.BOOLEAN] = ListFieldTypeEnum.BOOLEAN
        result[FormFieldTypeEnum.DATE] = ListFieldTypeEnum.DATE
        result[FormFieldTypeEnum.DATETIME] = ListFieldTypeEnum.DATETIME
        result[FormFieldTypeEnum.TIME] = ListFieldTypeEnum.TIME
        result[FormFieldTypeEnum.EMAIL] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.FILE] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.NUMBER] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.INTEGER] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.FLOAT] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.PASSWORD] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.TELEPHONE] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.STRING] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.TEXT] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.URL] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.UUID] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.IPV4] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.IPV6] = ListFieldTypeEnum.STRING
        result[FormFieldTypeEnum.OBJECT] = ListFieldTypeEnum.OBJECT
        return result

    def _load_base_url(self):
        """
        loads admin base url from `admin` config store.

        :rtype: str
        """

        url = config_services.get_active('admin', 'url')
        if not url.endswith('/'):
            url = f'{url}/'

        return url

    def _remove_from_pages(self, register_name):
        """
        removes the admin page with given register name from admin pages.

        it returns the removed admin page.

        :param str register_name: register name of the admin page to be removed.

        :rtype: pyrin.admin.interface.AbstractAdminPage
        """

        instance = self._admin_pages.pop(register_name, None)
        if instance is not None:
            return self._remove_from_entities(instance.get_entity())

        return instance

    def _remove_from_entities(self, entity):
        """
        removes the admin page with given entity from admin entities.

        it returns the removed admin page.

        :param type[pyrin.database.model.base.BaseEntity] entity: the entity class of
                                                                  admin page to be removed.

        :rtype: pyrin.admin.interface.AbstractAdminPage
        """

        instance = self._admin_entities.pop(entity, None)
        if instance is not None:
            return self._remove_from_pages(instance.get_register_name())

        return None

    def _get_admin_page(self, register_name):
        """
        gets the admin page with given register name.

        :param str register_name: register name of admin page to be get.
                                  this name is case-insensitive.

        :raises AdminPageNotFoundError: admin page not found error.

        :rtype: pyrin.admin.interface.AbstractAdminPage
        """

        name = str(register_name).lower()
        if name not in self._admin_pages:
            raise AdminPageNotFoundError(_('Admin page [{name}] not found.')
                                         .format(name=name))

        return self._admin_pages.get(name)

    def try_get_admin_page(self, entity):
        """
        gets the admin page for given entity class.

        it returns None if admin page does not exist.

        :param type[pyrin.database.model.base.BaseEntity] entity: the entity class of
                                                                  admin page to be get.

        :rtype: pyrin.admin.interface.AbstractAdminPage
        """

        return self._admin_entities.get(entity)

    def is_admin_enabled(self):
        """
        gets a value indicating that admin api is enabled.

        :rtype: bool
        """

        return config_services.get_active('admin', 'enabled')

    def has_admin(self, entity):
        """
        gets a value indicating that given entity class has admin page.

        :param type[pyrin.database.model.base.BaseEntity] entity: entity class.

        :rtype: bool
        """

        return entity in self._admin_entities

    def get_admin_base_url(self):
        """
        gets admin base url.

        :rtype: str
        """

        return self._base_url

    def get_admin_configurations(self):
        """
        gets the admin api configurations.

        :returns: dict(bool enabled: enable admin api,
                       bool authenticated: admin api access type,
                       str url: admin api base url)
        :rtype: dict
        """

        return config_services.get_active_section('admin')

    def get_default_category(self):
        """
        gets the default category to be used for admin pages without category.

        :rtype: str
        """

        category = config_services.get_active('admin', 'default_category')
        return category.upper()

    def register(self, instance, **options):
        """
        registers the provided instance into available admin pages.

        :param pyrin.admin.interface.AbstractAdminPage instance: admin page instance.

        :keyword bool replace: specifies that if another admin page with the same name
                               or the same entity exists, replace it.
                               defaults to False if not provided and raises an error.

        :raises InvalidAdminPageTypeError: invalid admin page type error.
        :raises DuplicatedAdminPageError: duplicated admin page error.
        """

        if not isinstance(instance, AbstractAdminPage):
            raise InvalidAdminPageTypeError('Input parameter [{admin}] is '
                                            'not an instance of [{base}].'
                                            .format(admin=instance, base=AbstractAdminPage))

        replace = options.get('replace', False)
        if instance.get_register_name() in self._admin_pages:
            if replace is not True:
                raise DuplicatedAdminPageError('There is another registered admin page '
                                               'with register name [{name}] but "replace" '
                                               'option is not set, so admin page [{instance}] '
                                               'could not be registered.'
                                               .format(name=instance.get_register_name(),
                                                       instance=instance))

            self._remove_from_pages(instance.get_register_name())

        if instance.get_entity() in self._admin_entities:
            if replace is not True:
                raise DuplicatedAdminPageError('There is another registered admin page '
                                               'for entity [{entity}] but "replace" '
                                               'option is not set, so admin page [{instance}] '
                                               'could not be registered.'
                                               .format(entity=instance.get_entity(),
                                                       instance=instance))

            self._remove_from_entities(instance.get_entity())

        self._admin_pages[instance.get_register_name()] = instance
        self._admin_entities[instance.get_entity()] = instance

    def get(self, register_name, pk):
        """
        gets an entity with given primary key.

        :param str register_name: register name of admin page.
        :param object pk: primary key of entity to be get.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        :raises EntityNotFoundError: entity not found error.

        :rtype: pyrin.database.model.base.BaseEntity
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_get_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow get operation.')
                                                .format(name=admin.get_register_name()))

        return admin.get(pk)

    def find(self, register_name, **filters):
        """
        performs find on given admin page and returns the result.

        :param str register_name: register name of admin page.

        :keyword **filters: all filters to be passed to related admin page.

        :rtype: list[ROW_RESULT]
        """

        admin = self._get_admin_page(register_name)
        return admin.find(**filters)

    def create(self, register_name, **data):
        """
        performs create on given admin page.

        :param str register_name: register name of admin page.

        :keyword **data: all data to be passed to related admin page for data creation.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.

        :rtype: object
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_create_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow create operation.')
                                                .format(name=admin.get_register_name()))

        return admin.create(**data)

    def update(self, register_name, pk, **data):
        """
        performs update on given admin page.

        :param str register_name: register name of admin page.
        :param object pk: entity primary key to be updated.

        :keyword **data: all data to be passed to related admin page for data creation.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        :raises EntityNotFoundError: entity not found error.
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_update_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow update operation.')
                                                .format(name=admin.get_register_name()))

        return admin.update(pk, **data)

    def remove(self, register_name, pk):
        """
        performs remove on given admin page.

        :param str register_name: register name of admin page.
        :param object pk: entity primary key to be removed.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_remove_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow remove operation.')
                                                .format(name=admin.get_register_name()))

        return admin.remove(pk)

    def remove_bulk(self, register_name, pk):
        """
        performs remove bulk on given admin page.

        :param str register_name: register name of admin page.
        :param object | list[object] pk: entity primary keys to be removed.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_remove_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow remove operation.')
                                                .format(name=admin.get_register_name()))

        return admin.remove_bulk(pk)

    def remove_all(self, register_name):
        """
        performs remove all on given admin page.

        :param str register_name: register name of admin page.

        :raises AdminOperationNotAllowedError: admin operation not allowed error.
        """

        admin = self._get_admin_page(register_name)
        if not admin.has_remove_all_permission():
            raise AdminOperationNotAllowedError(_('Admin page [{name}] does '
                                                  'not allow remove all operation.')
                                                .format(name=admin.get_register_name()))

        return admin.remove_all()

    def populate_main_metadata(self):
        """
        populates all admin pages main metadata.
        """

        metadata = dict()
        for name, admin in self._admin_pages.items():
            pages = metadata.setdefault(admin.get_category(), [])
            pages.append(admin.get_main_metadata())

        result = list()
        sorted_categories = sorted(metadata.keys())
        for category in sorted_categories:
            pages = metadata.get(category)
            sorted_pages = dict_utils.extended_sort(pages, 'plural_name')
            single_category = dict()
            single_category[category] = sorted_pages
            result.append(single_category)

        self._admin_metadata = tuple(result)

    def get_main_metadata(self):
        """
        gets all admin pages main metadata.

        :raises AdminPagesHaveNotLoadedError: admin pages have not loaded error.

        :returns: dict(list pages,
                       dict configs)
        :rtype: dict
        """

        if self._admin_metadata is None:
            raise AdminPagesHaveNotLoadedError('Admin pages have not loaded yet.')

        result = dict(pages=list(self._admin_metadata), configs=self.get_configs())
        return result

    def get_find_metadata(self, register_name):
        """
        gets the find metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        admin = self._get_admin_page(register_name)
        return admin.get_find_metadata()

    def get_create_metadata(self, register_name):
        """
        gets the create metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        admin = self._get_admin_page(register_name)
        return admin.get_create_metadata()

    def get_update_metadata(self, register_name):
        """
        gets the update metadata for given admin page.

        :param str register_name: register name of admin page.

        :rtype: dict
        """

        admin = self._get_admin_page(register_name)
        return admin.get_update_metadata()

    def register_name_for(self, entity):
        """
        gets the admin page register name for given entity class.

        it raises an error if the given entity does not have an admin page.

        :param type[pyrin.database.model.base.BaseEntity] entity: the entity class of
                                                                  admin page to get its
                                                                  register name.

        :raises AdminPageNotFoundError: admin page not found error.

        :rtype: str
        """

        admin = self.try_get_admin_page(entity)
        if admin is None:
            raise AdminPageNotFoundError(_('Entity [{entity}] does not have an admin page.')
                                         .format(entity=entity))

        return admin.get_register_name()

    def url_for(self, register_name):
        """
        gets the base url for given admin page.

        :param str register_name: admin page register name.

        :rtype: str
        """

        return f'{self.get_admin_base_url()}{register_name.lower()}/'

    def get_list_field_type(self, form_field_type):
        """
        gets the equivalent list field type for given form field type.

        it may return None.

        :param str form_field_type: form field type to get its list field type.
        :enum form_field_type:
            BOOLEAN = 'boolean'
            DATE = 'date'
            DATETIME = 'datetime'
            TIME = 'time'
            EMAIL = 'email'
            FILE = 'file'
            NUMBER = 'number'
            PASSWORD = 'password'
            TELEPHONE = 'telephone'
            STRING = 'string'
            TEXT = 'text'
            URL = 'url'
            UUID = 'uuid'
            IPV4 = 'ipv4'
            IPV6 = 'ipv6'
            OBJECT = 'object'

        :rtype: str
        """

        return self._type_map.get(form_field_type)

    def populate_caches(self):
        """
        populates required caches of all registered admin pages.

        :returns: count of registered admin pages
        :rtype: int
        """

        for admin in self._admin_pages.values():
            admin.populate_caches()

        return len(self._admin_pages)

    def get_configs(self):
        """
        gets the required configs of admin api.

        :returns: dict(str panel_name,
                       str page_key,
                       str page_size_key,
                       str ordering_key,
                       str locale_key,
                       str timezone_key,
                       str search_param,
                       str hidden_pk_name)
        :rtype: dict
        """

        if self._configs is not None:
            return deepcopy(self._configs)

        panel_name = config_services.get_active('admin', 'panel_name')
        search_param = config_services.get_active('admin', 'search_param')
        hidden_pk_name = config_services.get_active('admin', 'hidden_pk_name')
        page_key, page_size_key = paging_services.get_paging_param_names()
        ordering_key = database_services.get_ordering_key()
        locale_key = locale_services.get_locale_key()
        timezone_key = datetime_services.get_timezone_key()
        result = dict(panel_name=panel_name, page_key=page_key,
                      page_size_key=page_size_key, ordering_key=ordering_key,
                      locale_key=locale_key, timezone_key=timezone_key,
                      search_param=search_param, hidden_pk_name=hidden_pk_name)

        self._configs = deepcopy(result)
        return result

    def login(self, username, password, **options):
        """
        logs in an internal user with given info into admin panel.

        :param str username: username.
        :param str password: password.

        :raises ProvidedUsernameOrPasswordAreIncorrect: provided username or
                                                        password are incorrect.

        :returns: dict(str access_token)
        :rtype: dict
        """

        return authentication_services.login(username, password,
                                             InternalAuthenticatorEnum.ADMIN,
                                             **options)



# -*- coding: utf-8 -*-
"""
users internal package.
"""

from pyrin.packaging.base import Package


class InternalUsersPackage(Package):
    """
    internal users package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'users.internal.component'
    DEPENDS = ['pyrin.admin',
               'pyrin.validator',
               'pyrin.database.model']



# -*- coding: utf-8 -*-
"""
caching models module.
"""

from sqlalchemy import LargeBinary

from pyrin.database.model.declarative import CoreEntity
from pyrin.database.model.mixin import CreateHistoryMixin
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.orm.sql.schema.columns import AutoPKColumn, StringColumn, BigIntegerColumn


class CacheItemBaseEntity(CoreEntity):
    """
    cache item base entity class.
    """

    _table = 'cache_item'

    id = AutoPKColumn(name='id')


class CacheItemEntity(CacheItemBaseEntity, CreateHistoryMixin):
    """
    cache item entity class.
    """

    _extend_existing = True

    cache_name = StringColumn(name='cache_name', max_length=30, nullable=False)
    shard_name = StringColumn(name='shard_name', max_length=30)
    version = StringColumn(name='version', max_length=20, nullable=False)
    key = BigIntegerColumn(name='key', nullable=False)
    item = CoreColumn(name='item', type_=LargeBinary, nullable=False)



# -*- coding: utf-8 -*-
"""
admin component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.admin import AdminPackage
from pyrin.admin.manager import AdminManager


@component(AdminPackage.COMPONENT_NAME)
class AdminComponent(Component, AdminManager):
    """
    admin component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
router handlers exceptions module.
"""

from pyrin.security.exceptions import AuthenticationFailedError
from pyrin.processor.request.wrappers.exceptions import BadRequestError
from pyrin.core.exceptions import CoreException, CoreBusinessException, CoreNotFoundError


class RouterHandlerException(CoreException):
    """
    router handler exception.
    """
    pass


class RouterHandlerBusinessException(CoreBusinessException,
                                     RouterHandlerException):
    """
    router handler business exception.
    """
    pass


class InvalidViewFunctionTypeError(RouterHandlerException):
    """
    invalid view function type error.
    """
    pass


class MaxContentLengthLimitMismatchError(RouterHandlerException):
    """
    max content length limit mismatch error.
    """
    pass


class FreshAuthenticationRequiredError(AuthenticationFailedError,
                                       RouterHandlerBusinessException):
    """
    fresh authentication required error.
    """
    pass


class PermissionTypeError(RouterHandlerException):
    """
    permission type error.
    """
    pass


class InvalidResultSchemaTypeError(RouterHandlerException):
    """
    invalid result schema type error.
    """
    pass


class RouteIsNotBoundedError(RouterHandlerException):
    """
    route is not bounded error.
    """
    pass


class RouteIsNotBoundedToMapError(RouterHandlerException):
    """
    route is not bounded to map error.
    """
    pass


class InvalidResponseStatusCodeError(RouterHandlerException):
    """
    invalid response status code error.
    """
    pass


class ViewFunctionRequiredParamsError(BadRequestError,
                                      RouterHandlerBusinessException):
    """
    view function required params error.
    """
    pass


class InvalidRequestLimitError(RouterHandlerException):
    """
    invalid request limit error.
    """
    pass


class InvalidLifetimeError(RouterHandlerException):
    """
    invalid lifetime error.
    """
    pass


class RequestLimitOrLifetimeRequiredError(RouterHandlerException):
    """
    request limit or lifetime required error.
    """
    pass


class URLNotFoundError(CoreNotFoundError, RouterHandlerBusinessException):
    """
    url not found error.
    """
    pass


class CouldNotFindRelevantAuthenticatorError(RouterHandlerException):
    """
    could not find relevant authenticator error.
    """
    pass



# -*- coding: utf-8 -*-
"""
application services module.
"""

from pyrin.application.container import _get_app


def add_context(key, value, **options):
    """
    adds the given key and it's value into the application context.

    :param str key: related key for storing application context.
    :param object value: related value for storing in application context.

    :keyword bool replace: specifies that if there is already a value with
                           the same key in application context, it should be updated
                           with new value, otherwise raise an error. defaults to False.

    :raises DuplicateContextKeyError: duplicate context key error.
    """

    get_current_app().add_context(key, value, **options)


def get_context(key, **options):
    """
    gets the application context value that belongs to given key.

    :param str key: key for requested application context.

    :keyword object default: default value to be returned if key is not available.
                             otherwise, it raises an error if key is not available.

    :raises ContextAttributeError: context attribute error.

    :returns: related value to given key.
    """

    return get_current_app().get_context(key, **options)


def register_component(component, **options):
    """
    registers given application component or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's id is already available
    in registered components.

    :param Component component: component instance.

    :keyword bool replace: specifies that if there is another registered
                           component with the same id, replace it with the new one.
                           otherwise raise an error. defaults to False.

    :raises InvalidComponentTypeError: invalid component type error.
    :raises InvalidComponentIDError: invalid component id error.
    :raises DuplicateComponentIDError: duplicate component id error.
    """

    get_current_app().register_component(component, **options)


def remove_component(component_id):
    """
    removes application component with given id.

    :param tuple[str, object] component_id: component id to be removed.

    :raises ComponentAttributeError: component attribute error.
    """

    get_current_app().remove_component(component_id)


def get_component(component_name, **options):
    """
    gets the specified application component.

    :param str component_name: component name.

    :keyword object component_custom_key: component custom key.
                                          if not provided, tries to get it from
                                          request object, if not found,
                                          `DEFAULT_COMPONENT_KEY` will be used.

    :raises InvalidComponentNameError: invalid component name error.
    :raises ComponentAttributeError: component attribute error.

    :rtype: Component
    """

    return get_current_app().get_component(component_name, **options)


def register_error_handler(code_or_exception, func):
    """
    registers the given function as an error handler for given code or exception type.

    :param int | type[Exception] code_or_exception: code or exception to
                                                    register error handler for.

    :param function func: function to register it as an error handler.
    """

    get_current_app().register_error_handler(code_or_exception, func)


def register_before_request_handler(func):
    """
    registers the given function into application before request handlers.

    :param function func: function to register it into before request handlers.
    """

    get_current_app().before_request(func)


def register_after_request_handler(func):
    """
    registers the given function into application after request handlers.

    :param function func: function to register it into after request handlers.
    """

    get_current_app().after_request(func)


def register_teardown_request_handler(func):
    """
    registers the given function into application teardown request handlers.

    teardown request handlers should not return any value
    and also should not raise any exception.

    :param function func: function to register it into teardown request handlers.
    """

    get_current_app().teardown_request(func)


def add_url_rule(rule, endpoint=None, view_func=None,
                 provide_automatic_options=None, **options):
    """
    connects a url rule. the provided `view_func` will be registered with the endpoint.

    if there is another rule with the same url and http methods and `replace=True`
    option is provided, it will be replaced. otherwise an error will be raised.

    a note about endpoint:

    pyrin will handle endpoint generation on its own.
    so it is recommended not to provide endpoint to this method.
    this is required to be able to handle uniqueness of endpoints and managing them.
    despite flask, pyrin will not require you to define view functions with unique names.
    you could define view functions with the same name in different modules. but to
    ensure the uniqueness of endpoints, pyrin will use the fully qualified name
    of function as endpoint. for example: `pyrin.api.services.create_route`.
    so you could figure out endpoint for any view function to use it in places
    like `url_for` method.

    pyrin handles endpoints this way to achieve these two important features:

    1. dismissal of the need for view function name uniqueness.
       when application grows, it's nonsense to be forced to have
       unique view function names at application level.

    2. managing routes with the same url and http methods, and informing
       the developer about them to prevent accidental replacements. and also
       providing a way to replace a route by another route with the same url
       and http methods if that is what developer actually wants.
       when application grows, it becomes a point of failure when you have no
       idea that you've registered a similar route in multiple places and only
       one of them will be get called based on registration order.

    :param str rule: the url rule as string.

    :param str endpoint: the endpoint of the route.
                         by default, it is the fully
                         qualified name of the view function.

    :param function view_func: the function to call when serving a request to the
                               provided endpoint.

    :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                           added automatically.
                                           this can also be controlled by setting the
                                           `view_func.provide_automatic_options = False`
                                           before adding the rule.

    :keyword str | tuple[str] methods: http methods that this rule should handle.
                                       if not provided, defaults to `GET`.

    :keyword PermissionBase | tuple[PermissionBase] permissions: all required permissions
                                                                 for accessing this route.

    :keyword bool authenticated: specifies that this route could not be accessed
                                 if the requester has not been authenticated.
                                 defaults to False if not provided.

    :keyword str authenticator: the authenticator name to be used for this route.
                                if not provided, it will be get from rule based
                                authenticators if possible. otherwise the
                                `default_authenticator` config will be used.
                                if no default is set in `authentication` config
                                store, it raises an error.
                                it is only used if this route has `authenticated=True`.

    :keyword bool fresh_auth: specifies that this route could not be accessed
                              if the requester has not a fresh authentication.
                              fresh authentication means an authentication that
                              has been done by providing user credentials to
                              server. defaults to False if not provided.

    :keyword bool replace: specifies that this route must replace any existing
                           route with the same url and http methods or raise
                           an error if not provided. defaults to False.

    :keyword dict defaults: an optional dict with defaults for other rules with the
                            same endpoint. this is a bit tricky but useful if you
                            want to have unique urls.

    :keyword str subdomain: the subdomain rule string for this rule. If not specified the
                            rule only matches for the `default_subdomain` of the map. if
                            the map is not bound to a subdomain this feature is disabled.

    :keyword bool strict_slashes: override the `Map` setting for `strict_slashes` only for
                                  this rule. if not specified the `Map` setting is used.

    :keyword bool merge_slashes: override `Map.merge_slashes` for this rule.

    :keyword bool build_only: set this to True and the rule will never match but will
                              create a url that can be build. this is useful if you have
                              resources on a subdomain or folder that are not handled by
                              the WSGI application (like static data)

    :keyword str | callable redirect_to: if given this must be either a string
                                         or callable. in case of a callable it's
                                         called with the url adapter that
                                         triggered the match and the values
                                         of the url as keyword arguments and has
                                         to return the target for the redirect,
                                         otherwise it has to be a string with
                                         placeholders in rule syntax.

    :keyword bool alias: if enabled this rule serves as an alias for another rule with
                         the same endpoint and arguments.

    :keyword str host: if provided and the url map has host matching enabled this can be
                       used to provide a match rule for the whole host. this also means
                       that the subdomain feature is disabled.

    :keyword bool websocket: if set to True, this rule is only matches for
                             websocket (`ws://`, `wss://`) requests. by default,
                             rules will only match for http requests.
                             defaults to False if not provided.

    :keyword int max_content_length: max content length that this route could handle,
                                     in bytes. if not provided, it will be set to
                                     `restricted_max_content_length` api config key.
                                     note that this value should be lesser than or equal
                                     to `max_content_length` api config key, otherwise
                                     it will cause an error.

    :keyword int status_code: status code to be returned on successful responses.
                              defaults to corresponding status code of request's
                              http method if not provided.

    :note status_code: it could be a value from `InformationResponseCodeEnum`
                       or `SuccessfulResponseCodeEnum` or `RedirectionResponseCodeEnum`.

    :keyword bool strict_status: specifies that it should only consider
                                 the status code as processed if it is from
                                 `InformationResponseCodeEnum` or
                                 `SuccessfulResponseCodeEnum` or
                                 `RedirectionResponseCodeEnum` values. otherwise
                                 all codes from `INFORMATION_CODE_MIN`
                                 to `INFORMATION_CODE_MAX` or from
                                 `SUCCESS_CODE_MIN` to `SUCCESS_CODE_MAX`
                                 or from `REDIRECTION_CODE_MIN` to
                                 `REDIRECTION_CODE_MAX` will be considered
                                 as processed. defaults to True if not provided.

    :keyword str | list[str] environments: a list of all environments that this
                                           route must be exposed on them.
                                           the values could be from all available
                                           environments in environments config store.
                                           for example: `production`, `development`.
                                           if not provided, the route will be exposed
                                           on all environments.

    :keyword ResultSchema | type[ResultSchema] result_schema: result schema to be used
                                                              to filter results. it could
                                                              be an instance or a type
                                                              of `ResultSchema` class.

    :keyword bool indexed: specifies that list results must
                           include an extra field as row index.
                           the name of the index field and the initial value
                           of index could be provided by `index_name` and
                           `start_index` respectively. `indexed` keyword has
                           only effect if the returning result contains a list
                           of objects.

    :keyword str index_name: name of the extra field to contain
                             the row index of each result. if not provided
                             defaults to `row_num` value.

    :keyword int start_index: the initial value of row index. if not
                              provided, starts from 1.

    :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                  which has `allow_read=False` or its name
                                                  starts with underscore `_`, should not
                                                  be included in result dict. defaults to
                                                  `SECURE_TRUE` if not provided. it will
                                                  be used only for entity conversion.
                                                  this value will override the
                                                  corresponding value of `result_schema`
                                                  if provided.

    :keyword int depth: a value indicating the depth for conversion.
                        for example if entity A has a relationship with
                        entity B and there is a list of B in A, if `depth=0`
                        is provided, then just columns of A will be available
                        in result dict, but if `depth=1` is provided, then all
                        B entities in A will also be included in the result dict.
                        actually, `depth` specifies that relationships in an
                        entity should be followed by how much depth.
                        note that, if `columns` is also provided, it is required to
                        specify relationship property names in provided columns.
                        otherwise they won't be included even if `depth` is provided.
                        defaults to `default_depth` value of database config store.
                        please be careful on increasing `depth`, it could fail
                        application if set to higher values. choose it wisely.
                        normally the maximum acceptable `depth` would be 2 or 3.
                        there is a hard limit for max valid `depth` which is set
                        in `ConverterMixin.MAX_DEPTH` class variable. providing higher
                        `depth` value than this limit, will cause an error.
                        it will be used only for entity conversion.
                        this value will override the corresponding value of
                        `result_schema` if provided.

    :keyword bool no_cache: a value indicating that the response returning from this route
                            must have a `Cache-Control: no-cache` header. this header will
                            be automatically added. defaults to False if not provided.

    :keyword int request_limit: number of allowed requests to this
                                route before it unregisters itself.
                                defaults to None if not Provided.

    :keyword int lifetime: number of seconds in which this route must remain
                           responsive after initial registration. after this
                           period, the route will unregister itself.
                           defaults to None if not provided.

    :note request_limit, lifetime: if both of these values are provided, this
                                   route will unregister itself if any of
                                   these two conditions are met.

    :keyword bool paged: specifies that this route should return paginated results.
                         defaults to False if not provided.

    :keyword int page_size: default page size for this route.
                            defaults to `default_page_size` from
                            `database` config store if not provided.

    :keyword int max_page_size: maximum page size that client is allowed
                                to request for this route. defaults to
                                `max_page_size` from `database` configs store
                                if not provided.

    :keyword bool cors_enabled: specifies that cross origin resource sharing is enabled.
                                if not provided, it will be get from cors config store.

    :keyword bool cors_always_send: specifies that cors headers must be included in
                                    response even if the request does not have origin header.
                                    if not provided, it will be get from cors config store.

    :keyword list[str] cors_allowed_origins: a list of extra allowed origins to be used
                                             in conjunction with default allowed ones.

    :keyword list[str] cors_exposed_headers: extra exposed headers to be combined
                                             with default ones.

    :keyword list[str] cors_allowed_headers: extra allowed headers to be combined
                                             with default ones.

    :keyword bool cors_allow_credentials: specifies that browsers are allowed to pass
                                          response headers to front-end javascript code
                                          if the route is authenticated.
                                          if not provided, it will be get from cors config
                                          store.

    :keyword int cors_max_age: maximum number of seconds to cache results.
                               if not provided, it will be get from cors config store.

    :keyword bool swagger: specifies that this route must be exposed on swagger.
                           defaults to False if not provided.

    :keyword bool ordered: specifies that this route provides ordered results.
                           this is a flag to be used by swagger package to add
                           `order_by` keyword into parameters.
                           defaults to False if not provided.

    :raises DuplicateRouteURLError: duplicate route url error.
    :raises OverwritingEndpointIsNotAllowedError: overwriting endpoint is not allowed error.
    :raises PageSizeLimitError: page size limit error.
    :raises MaxContentLengthLimitMismatchError: max content length limit mismatch error.
    :raises InvalidViewFunctionTypeError: invalid view function type error.
    :raises InvalidResultSchemaTypeError: invalid result schema type error.
    :raises InvalidResponseStatusCodeError: invalid response status code error.
    """

    get_current_app().add_url_rule(rule, endpoint, view_func,
                                   provide_automatic_options, **options)


def generate_endpoint(func, **options):
    """
    generates endpoint for given function.

    pyrin will assume endpoint as function's fully qualified name.
    this method could be overridden in subclasses to change the endpoint generation.

    :param function func: function to generate endpoint for it.

    :rtype: str
    """

    return get_current_app().generate_endpoint(func, **options)


def register_route_factory(factory):
    """
    registers a route factory as application url rule class.

    :param callable factory: route factory.
                             it could be a class or a factory method.

    :raises InvalidRouteFactoryTypeError: invalid route factory type error.
    """

    get_current_app().register_route_factory(factory)


def get_current_route_factory():
    """
    gets current route factory in use.

    :rtype: callable
    """

    return get_current_app().get_current_route_factory()


def get_default_settings_path():
    """
    gets the pyrin default settings path.

    :rtype: str
    """

    return get_current_app().get_default_settings_path()


def get_settings_path():
    """
    gets the application settings path.

    :rtype: str
    """

    return get_current_app().get_settings_path()


def get_migrations_path():
    """
    gets the application migrations path.

    :rtype: str
    """

    return get_current_app().get_migrations_path()


def get_locale_path():
    """
    gets the application locale path.

    :rtype: str
    """

    return get_current_app().get_locale_path()


def get_application_main_package_path():
    """
    gets the application main package path.

    :rtype: str
    """

    return get_current_app().get_application_main_package_path()


def get_pyrin_root_path():
    """
    gets pyrin root path in which pyrin package is located.

    :rtype: str
    """

    return get_current_app().get_pyrin_root_path()


def get_application_root_path():
    """
    gets the application root path in which application package is located.

    :rtype: str
    """

    return get_current_app().get_application_root_path()


def get_pyrin_main_package_path():
    """
    gets pyrin main package path.

    :rtype: str
    """

    return get_current_app().get_pyrin_main_package_path()


def get_working_directory():
    """
    gets working directory path of application.

    working directory is where the root application and test package are resided.
    this is required when application starts from any of test applications.
    then we should move root path up, to the correct root to be able to
    include real application packages too.
    if the application has been started from real application, this method
    returns the same result as `get_application_root_path()` method.

    :rtype: str
    """

    return get_current_app().get_working_directory()


def get_configs():
    """
    gets a shallow copy of application's configuration dictionary.

    :rtype: dict
    """

    return get_current_app().get_configs()


def configure(config_store):
    """
    configures the application with given dict.

    all keys will be converted to uppercase for flask compatibility.

    :param dict config_store: a dictionary containing configuration key/values.
    """

    get_current_app().configure(config_store)


def load_configs(**options):
    """
    loads all configurations related to application package.

    normally, you should not call this method manually.
    """

    get_current_app().load_configs(**options)


def get_current_app():
    """
    gets the instance of current running application.

    :rtype: Application
    """

    return _get_app()


def register_hook(instance):
    """
    registers the given instance into application hooks.

    :param ApplicationHookBase instance: application hook instance to be registered.

    :raises InvalidApplicationHookTypeError: invalid application hook type error.
    """

    get_current_app().register_hook(instance)


def is_scripting_mode():
    """
    gets a value indicating that application has been started in scripting mode.

    some application hooks will not fire in this mode. like 'before_application_run'.
    """

    return get_current_app().is_scripting_mode()


def get_application_name():
    """
    gets the application name.

    it is actually the application main package name.
    first, it gets the import name of application if available.
    otherwise gets the first part of application package name

    :rtype: str
    """

    return get_current_app().get_application_name()


def get_class_name():
    """
    gets the application class name.

    it is required for template generation.

    :rtype: str
    """

    return get_current_app().get_class_name()


def get_module_name():
    """
    gets the application module name.

    it is required for template generation.

    :rtype: str
    """

    return get_current_app().get_module_name()


def get_application_version():
    """
    gets application version.

    :rtype: str
    """

    return get_current_app().get_application_version()


def get_pyrin_version():
    """
    gets pyrin version.

    :rtype: str
    """

    return get_current_app().get_pyrin_version()


def got_first_request():
    """
    gets a value indicating that application has been served at least one request.

    :rtype: bool
    """

    return get_current_app().got_first_request


def get_current_url_adapter():
    """
    gets url adapter of current request.

    if request context is not present, this method raises an error.

    :rtype: MapAdapter
    """

    return get_current_app().get_current_url_adapter()



# -*- coding: utf-8 -*-
"""
users internal validators module.
"""

from pyrin.admin.enumerations import FormFieldTypeEnum
from pyrin.users.internal.models import InternalUserEntity
from pyrin.validator.decorators import validator
from pyrin.validator.handlers.string import StringValidator


@validator(InternalUserEntity, InternalUserEntity.password_hash, name='password')
@validator(InternalUserEntity, InternalUserEntity.password_hash, name='confirm_password')
class PasswordValidator(StringValidator):
    """
    password validator class.
    """

    default_minimum_length = 8
    default_maximum_length = 50
    default_form_field_type = FormFieldTypeEnum.PASSWORD



# -*- coding: utf-8 -*-
"""
authorization exceptions module.
"""

from pyrin.core.exceptions import CoreBusinessException, CoreException
from pyrin.security.exceptions import AuthenticationFailedError, AuthorizationFailedError


class AuthorizationManagerException(CoreException):
    """
    authorization manager exception.
    """
    pass


class AuthorizationManagerBusinessException(CoreBusinessException,
                                            AuthorizationManagerException):
    """
    authorization manager business exception.
    """
    pass


class UserNotAuthenticatedError(AuthenticationFailedError,
                                AuthorizationFailedError):
    """
    user not authenticated error.
    """
    pass


class PermissionDeniedError(AuthorizationFailedError):
    """
    permission denied error.
    """
    pass


class InvalidAuthorizerTypeError(AuthorizationManagerException):
    """
    invalid authorizer type error.
    """
    pass


class DuplicatedAuthorizerError(AuthorizationManagerException):
    """
    duplicated authorizer error.
    """
    pass


class AuthorizerNotFoundError(AuthorizationManagerException):
    """
    authorizer not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
admin security exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.security.exceptions import AuthorizationFailedError


class AdminSecurityException(CoreException):
    """
    admin security exception.
    """
    pass


class AdminSecurityBusinessException(CoreBusinessException,
                                     AdminSecurityException):
    """
    admin security business exception.
    """
    pass


class AdminAccessNotAllowedError(AuthorizationFailedError,
                                 AdminSecurityBusinessException):
    """
    admin access not allowed error.
    """
    pass



# -*- coding: utf-8 -*-
"""
authentication interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class AuthenticatorSingletonMeta(MultiSingletonMeta):
    """
    authenticator singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractAuthenticatorBase(CoreObject, metaclass=AuthenticatorSingletonMeta):
    """
    abstract authenticator base class.
    """

    @abstractmethod
    def authenticate(self, request, **options):
        """
        authenticates the user for given request.

        :param CoreRequest request: current request object.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def is_fresh(self):
        """
        gets a value indicating that current request has a fresh authentication.

        fresh authentication means an authentication which is done by providing
        user credentials to server.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def login(self, username, password, **options):
        """
        logs in a user with given info and stores/generates the relevant credentials.

        it may return the required credentials if they must be returned to client.

        :param str username: username.
        :param str password: password.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: required credentials.
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def logout(self, **options):
        """
        logouts the current user and clears its relevant credentials.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this authenticator.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
authorization handlers base module.
"""

from abc import abstractmethod

import pyrin.security.session.services as session_services
import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _, _n
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.authorization.interface import AbstractAuthorizerBase
from pyrin.security.exceptions import UserIsNotActiveError
from pyrin.security.authorization.handlers.exceptions import AuthorizerNameIsRequiredError
from pyrin.security.authorization.exceptions import UserNotAuthenticatedError, \
    PermissionDeniedError


class AuthorizerBase(AbstractAuthorizerBase):
    """
    authorizer base class.

    all application authorizers must be subclassed from this.
    """

    # each subclass must set an authorizer name in this attribute.
    _name = None

    def __init__(self, *args, **options):
        """
        initializes and instance of AuthorizerBase.

        :raises AuthorizerNameIsRequiredError: authorizer name is required error.
        """

        super().__init__()

        if not self._name or self._name.isspace():
            raise AuthorizerNameIsRequiredError('Authorizer [{instance}] does not '
                                                'have a name.'.format(instance=self))

    def _authorize_access(self, user, **options):
        """
        authorizes the given user for custom accesses.

        this method could be overridden in subclasses if required.
        it must raise an error if authorization failed.

        :param user: user identity to be checked if it is authorized.

        :keyword dict user_info: user info to be used to check for user.

        :raises AuthorizationFailedError: authorization failed error.
        """
        pass

    def _is_superuser(self, user, **options):
        """
        gets a value indicating that given user is superuser.

        this method could be overridden in subclasses to provide actual
        implementation for checking that a user is superuser.
        otherwise this method will always return False.

        :param user: user identity to be checked if it is superuser.

        :keyword dict user_info: user info to be used to check for user.

        :rtype: bool
        """

        return False

    def _is_active(self, user, **options):
        """
        gets a value indicating that the given user is active.

        this method could be overridden in subclasses to provide actual
        implementation for checking that a user is active.
        otherwise this method will always return True.

        :param user: user identity to be checked if it is active.

        :keyword dict user_info: user info to be used to check for user.

        :rtype: bool
        """

        return True

    def _authorize_permissions(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to be checked if it is authorized.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword dict user_info: user info to be used for authorization.

        :raises PermissionDeniedError: permission denied error.
        """

        permissions = misc_utils.make_iterable(permissions, tuple)
        if len(permissions) > 0:
            if self._has_permission(user, permissions, **options) is not True:
                message = _n('You do not have the required permission {permissions}',
                             'You do not have the required permissions {permissions}',
                             len(permissions))
                raise PermissionDeniedError(message.format(permissions=list(permissions)))

    def authorize(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.

        if user does not have each one of the specified
        permissions, an error will be raised.

        :param user: user identity to authorize permissions for.

        :param PermissionBase | list[PermissionBase] permissions: permissions to check
                                                                  for user authorization.

        :keyword dict user_info: user info to be used for authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        :raises PermissionDeniedError: permission denied error.
        """

        if user is None:
            raise UserNotAuthenticatedError(_('User has not been authenticated.'))

        if self._is_active(user, **options) is not True:
            raise UserIsNotActiveError(_('Your user is not active. If you think that this '
                                         'is a mistake, please contact the support team.'))

        if self._is_superuser(user, **options) is not True:
            self._authorize_access(user, **options)
            if permissions:
                self._authorize_permissions(user, permissions, **options)

    def is_superuser(self):
        """
        gets a value indicating that the current user is superuser.

        if you want to provide the actual implementation for checking that a
        user is superuser, you should implement the `_is_superuser` method.
        otherwise this method will always return False.

        :rtype: bool
        """

        user = session_services.get_current_user()
        user_info = session_services.get_current_user_info()
        return self._is_superuser(user, user_info=user_info)

    @abstractmethod
    def _has_permission(self, user, permissions, **options):
        """
        gets a value indicating that given user has the requested permissions.

        this method must be overridden in subclasses to perform a check on permissions.

        :param user: user identity to authorize permissions for.
        :param tuple[PermissionBase] permissions: permissions to check for user authorization.

        :keyword dict user_info: user info to be used for authorization.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @property
    def name(self):
        """
        gets the name of this authenticator.

        :rtype: str
        """

        return self._name



# -*- coding: utf-8 -*-
"""
utils dictionary module.
"""

from operator import itemgetter
from collections import OrderedDict

import pyrin.utils.string as string_utils

from pyrin.core.structs import DTO
from pyrin.utils.exceptions import InputNotCallableError, DictKeyPrefixIsNotProvidedError


def change_key_case(value, converter, **options):
    """
    returns a copy of input dict with all it's keys
    and nested keys cases modified using given converter.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to change it's keys cases.

    :param callable converter: a callable to use as case converter.
                               it should be a callable with a signature
                               similar to below example.
                               case_converter(input_dict).

    :raises InputNotCallableError: input not callable error.

    :rtype: dict
    """

    if not callable(converter):
        raise InputNotCallableError('Input parameter [{converter}] is not callable.'
                                    .format(converter=str(converter)))

    result_dict = DTO()
    for key, val in value.items():
        if isinstance(val, dict):
            val = change_key_case(val, converter, **options)
        result_dict[converter(key)] = val
    return result_dict


def make_key_upper(value):
    """
    returns a copy of input dict with all it's
    keys and nested keys cases made upper.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to make it's keys uppercase.

    :rtype: dict
    """

    return change_key_case(value, string_utils.upper)


def make_key_lower(value):
    """
    returns a copy of input dict with all it's
    keys and nested keys cases made lower.
    note that this only considers direct nested dicts, if there is a
    list that contains a dict as value, this method won't change that.

    :param dict value: dict to make it's keys lowercase.

    :rtype: dict
    """

    return change_key_case(value, string_utils.lower)


def remove_keys(value, prefix):
    """
    removes all keys from given dict which starting with given prefix and
    also those keys that their names are exactly like prefixed ones, but without prefix.
    for example: dict(name='some_name', value='value', __value='another_value')
    if prefix is double underscores `__` then this function will return a new dict
    result = dict(name='some_name') with just the name key.
    note that this function just loops through top level keys.

    :param dict value: dict to remove keys from it.
    :param str prefix: keys prefix that should be removed from dict.

    :raises DictKeyPrefixIsNotProvidedError: dict key prefix is not provided error.

    :rtype: dict
    """

    if prefix in (None, '') or prefix.isspace():
        raise DictKeyPrefixIsNotProvidedError('Dictionary key prefix must be provided '
                                              'to remove keys with such prefix.')

    prefix_length = len(prefix)
    prefixed_to_remove = [key for key in value.keys() if key.startswith(prefix)]
    original_to_remove = [key[prefix_length:] for key in prefixed_to_remove]
    prefixed_to_remove.extend(original_to_remove)
    result = DTO(**value)
    for key in prefixed_to_remove:
        result.pop(key, None)

    return result


def sort_by_value(value, reverse=False):
    """
    sorts a dictionary by its values.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: x[1], reverse=reverse)
    return OrderedDict(result)


def sort_by_key(value, reverse=False):
    """
    sorts a dictionary by its keys.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: x[0], reverse=reverse)
    return OrderedDict(result)


def sort_by_key_length(value, reverse=False):
    """
    sorts a dictionary by its keys length.

    note that keys must be strings.

    :param dict value: dict to be sorted.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: OrderedDict
    """

    result = sorted(value.items(), key=lambda x: len(x[0]), reverse=reverse)
    return OrderedDict(result)


def extended_sort(items, key, reverse=False):
    """
    sorts a list of dictionaries by values of specified key.

    note that the provided key must be present in all items and the value
    of this key must be comparable for all items otherwise it raises an error.

    :param list[dict] items: items to be sorted.
    :param object key: key of dict to sort list items by it.

    :param bool reverse: sort by descending order.
                         defaults to False if not provided.

    :rtype: list[dict]
    """

    return sorted(items, key=itemgetter(key), reverse=reverse)


def create_dict(items):
    """
    creates a dict with each key and value set to an item of given list.

    :param list items: items to create a dict from them.

    :rtype: dict
    """

    result = {name: name for name in items}
    return result


def create_options(items):
    """
    creates a tuple of dicts with each key and value set to an item of given list.

    :param list items: items to create a tuple of dicts from them.

    :rtype: tuple[dict]
    """

    results = []
    for item in items:
        option = dict(name=item, value=item)
        results.append(option)

    return tuple(results)



# -*- coding: utf-8 -*-
"""
admin hooks module.
"""

import pyrin.admin.services as admin_services

from pyrin.utils.custom_print import print_info
from pyrin.validator.auto.hooks import AutoValidatorHookBase
from pyrin.validator.auto.decorators import auto_validator_hook


@auto_validator_hook()
class AutoValidatorHook(AutoValidatorHookBase):
    """
    auto validator hook class.
    """

    def after_auto_validators_registered(self):
        """
        this method will be called after all application auto validators have been registered.
        """

        count = admin_services.populate_caches()
        admin_services.populate_main_metadata()
        if count > 0:
            print_info('Total of [{count}] admin pages registered.'.format(count=count))



# -*- coding: utf-8 -*-
"""
admin package.
"""

from pyrin.packaging.base import Package


class AdminPackage(Package):
    """
    admin package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'admin.component'
    CONFIG_STORE_NAMES = ['admin']
    DEPENDS = ['pyrin.api.router',
               'pyrin.configuration',
               'pyrin.validator.auto']



# -*- coding: utf-8 -*-
"""
admin enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class ListFieldTypeEnum(CoreEnum):
    """
    list field type enum.
    """

    BOOLEAN = 'boolean'
    NUMERIC = 'numeric'
    DATE = 'date'
    DATETIME = 'datetime'
    TIME = 'time'
    STRING = 'string'
    CURRENCY = 'currency'
    OBJECT = 'object'


class FormFieldTypeEnum(CoreEnum):
    """
    form field type enum.
    """

    BOOLEAN = 'boolean'
    DATE = 'date'
    DATETIME = 'datetime'
    TIME = 'time'
    EMAIL = 'email'
    FILE = 'file'
    NUMBER = 'number'
    INTEGER = 'integer'
    FLOAT = 'float'
    PASSWORD = 'password'
    TELEPHONE = 'telephone'
    STRING = 'string'
    TEXT = 'text'
    URL = 'url'
    UUID = 'uuid'
    IPV4 = 'ipv4'
    IPV6 = 'ipv6'
    OBJECT = 'object'



# -*- coding: utf-8 -*-
"""
utils sqlalchemy module.
"""

from sqlalchemy.sql import quoted_name
from sqlalchemy.engine import result_tuple
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import inspect, Table, asc, desc, func, CheckConstraint

import pyrin.utils.datetime as datetime_utils
import pyrin.utils.string as string_utils

from pyrin.core.globals import _
from pyrin.core.globals import LIST_TYPES
from pyrin.utils.exceptions import InvalidRowResultFieldsAndValuesError, \
    FieldsAndValuesCountMismatchError, CheckConstraintValuesRequiredError, \
    MultipleDeclarativeClassesFoundError, InvalidOrderingColumnError


LIKE_CHAR_COUNT_LIMIT = 20


def like_both(value, start='%', end='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to both ends.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    value = like_prefix(value, start)
    value = like_suffix(value, end)

    return value


def like_prefix(value, start='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to beginning.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{start}{value}'.format(start=start, value=value)


def like_suffix(value, end='%'):
    """
    gets a copy of string with `%` or couple of `_` values attached to end.

    it is to be used in like operator.

    :param str value: value to be processed.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{value}{end}'.format(value=value, end=end)


def _process_place_holder(value, count):
    """
    processes the value and generates a place holder with count of `_` chars.

    this value could be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    if value is None:
        return ''

    if count is None or count <= 0:
        return ''

    place_holder = None
    if count > LIKE_CHAR_COUNT_LIMIT:
        place_holder = '%'

    if place_holder is None:
        place_holder = '_' * count

    return place_holder


def like_exact_both(value, count):
    """
    gets a copy of string with `count` number of `_` attached to both ends.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_both(value, place_holder, place_holder)


def like_exact_prefix(value, count):
    """
    gets a copy of string with `count` number of `_` attached to beginning.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_prefix(value, place_holder)


def like_exact_suffix(value, count):
    """
    gets a copy of string with `count` number of `_` attached to end.

    it is to be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_suffix(value, place_holder)


def add_range_clause(clauses, column, value_lower, value_upper,
                     include_equal_to_lower=True,
                     include_equal_to_upper=True,
                     **options):
    """
    adds range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add range clause to it.
    :param CoreColumn column: entity column to add range clause for it.
    :param object value_lower: lower bound of range clause.
    :param object value_upper: upper bound of range clause.

    :param include_equal_to_lower: specifies that lower value
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper value
                                   should be considered in range.
                                   defaults to True if not provided.
    """

    if value_lower is None and value_upper is None:
        return

    if value_lower is not None and value_lower == value_upper:
        clauses.append(column == value_lower)
    else:
        if value_lower is not None:
            if include_equal_to_lower is True:
                clauses.append(column >= value_lower)
            else:
                clauses.append(column > value_lower)
        if value_upper is not None:
            if include_equal_to_upper is True:
                clauses.append(column <= value_upper)
            else:
                clauses.append(column < value_upper)


def add_datetime_range_clause(clauses, column,
                              value_lower, value_upper,
                              include_equal_to_lower=True,
                              include_equal_to_upper=True,
                              **options):
    """
    adds datetime range comparison into given clauses using specified inputs.

    if the inputs are date objects, they will be converted to datetime with client
    timezone and `consider_begin_of_day` and `consider_end_of_day` will also
    considered as True.

    :param list clauses: clause list to add datetime range clause to it.
    :param CoreColumn column: entity column to add datetime range clause for it.
    :param datetime | date value_lower: lower bound of datetime range clause.
    :param datetime | date value_upper: upper bound of datetime range clause.

    :param include_equal_to_lower: specifies that lower datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :keyword bool consider_begin_of_day: specifies that consider begin
                                         of day for lower datetime.
                                         defaults to False if not provided.

    :keyword bool consider_end_of_day: specifies that consider end
                                       of day for upper datetime.
                                       defaults to False if not provided.
    """

    value_lower, value_upper = datetime_utils.normalize_datetime_range(value_lower,
                                                                       value_upper,
                                                                       **options)

    add_range_clause(clauses, column,
                     value_lower, value_upper,
                     include_equal_to_lower,
                     include_equal_to_upper,
                     **options)


def add_string_range_clause(clauses, column, value_lower, value_upper,
                            include_equal_to_lower=True,
                            include_equal_to_upper=True,
                            **options):
    """
    adds string range comparison into given clauses using specified inputs.

    it coerces both values into string if they are not None.
    it also converts both values and the value of database column into lowercase.

    :param list clauses: clause list to add string range clause to it.
    :param CoreColumn column: entity column to add string range clause for it.
    :param object value_lower: lower bound of string range clause.
    :param object value_upper: upper bound of string range clause.

    :param include_equal_to_lower: specifies that lower value
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper value
                                   should be considered in range.
                                   defaults to True if not provided.
    """

    if value_lower is not None:
        value_lower = str(value_lower).lower()

    if value_upper is not None:
        value_upper = str(value_upper).lower()

    # swapping values in case of user mistake.
    if value_lower is not None and value_upper is not None:
        if value_lower > value_upper:
            value_lower, value_upper = value_upper, value_lower

    add_range_clause(clauses, func.lower(column),
                     value_lower, value_upper,
                     include_equal_to_lower,
                     include_equal_to_upper,
                     **options)


def add_comparison_clause(clauses, column, value, **options):
    """
    adds list or single comparison into clauses based on given value.

    if the value type is any of list, tuple or set, it generates an
    `in()` comparison, otherwise it generates a simple `==` comparison.

    :param list clauses: clause list to add comparison clause to it.
    :param CoreColumn column: entity column to add comparison clause for it.
    :param object | list[object] value: value to add comparison for it.
    """

    if value is not None:
        if isinstance(value, LIST_TYPES):
            clauses.append(column.in_(value))
        else:
            clauses.append(column == value)


def create_row_result(fields, values):
    """
    creates a row result object with given fields and values.

    this object type is returned by sqlalchemy `Query` when there
    is column names or multiple entities in query.

    :param list[str] fields: field names of the result object.

    :param list[object] values: values to be mapped to fields.
                                they must be in the same order as fields.

    :raises InvalidRowResultFieldsAndValuesError: invalid row result fields
                                                  and values error.

    :raises FieldsAndValuesCountMismatchError: fields and values count mismatch error.

    :rtype: ROW_RESULT
    """

    if fields is None or values is None:
        raise InvalidRowResultFieldsAndValuesError('Input parameters "fields" and '
                                                   '"values" must both be provided, '
                                                   'they could not be None.')

    if len(fields) != len(values):
        raise FieldsAndValuesCountMismatchError('The length of "fields" which is '
                                                '[{fields}] and "values" which is '
                                                '[{values}] does not match.'
                                                .format(fields=len(fields),
                                                        values=len(values)))
    result = result_tuple(fields)
    return result(values)


def check_constraint(column, values, **options):
    """
    generates a check constraint for given column and values.

    by default, it generates an `in` check, but this could be changed to
    `not in` by providing `use_in=False` in options.

    if the first item of values is a string, all values will be quoted.

    :param str column: column name to be used in check constraint.
    :param list | tuple values: values to be used in check constraint.

    :keyword bool use_in: specifies that it must generate an `in` check.
                          otherwise it generates a `not in` check.
                          defaults to True if not provided.

    :keyword **options: all other keyword arguments will be passed to
                        underlying `CheckConstraint` constructor.

    :raises CheckConstraintValuesRequiredError: check constraint values required error.

    :rtype: CheckConstraint
    """

    if values is None or len(values) <= 0:
        raise CheckConstraintValuesRequiredError('Values for generating a check '
                                                 'constraint must be provided.')

    converter = str
    is_string = isinstance(values[0], str)
    if is_string:
        converter = string_utils.quote
    use_in = options.pop('use_in', True)
    condition = 'in'
    if use_in is False:
        condition = 'not in'

    string_values = ', '.join(converter(item) for item in values)
    sql_text = '{column} {condition} ({values})'.format(column=quoted_name(column, True),
                                                        condition=condition,
                                                        values=string_values)
    options.update(sqltext=sql_text)
    return CheckConstraint(**options)


def range_check_constraint(column, min_value=None, max_value=None, **options):
    """
    generates a range check constraint for given column and values.

    if the values are string, they will be quoted.
    if only one value is provided, a max or min constraint will be generated.

    :param str column: column name to be used in check constraint.
    :param object min_value: min value to be used in check constraint.
    :param object max_value: max value to be used in check constraint.

    :keyword **options: all other keyword arguments will be passed to
                        underlying `CheckConstraint` constructor.

    :raises CheckConstraintValuesRequiredError: check constraint values required error.

    :rtype: CheckConstraint
    """

    if min_value is None and max_value is None:
        raise CheckConstraintValuesRequiredError('Values for generating a range check '
                                                 'constraint must be provided.')

    sql_text = None
    if min_value is not None and max_value is not None:
        sql_text = '{column} >= {min} and {column} <= {max}'
    elif min_value is not None:
        sql_text = '{column} >= {min}'
    else:
        sql_text = '{column} <= {max}'

    if isinstance(min_value, str):
        min_value = string_utils.quote(min_value)

    if isinstance(max_value, str):
        max_value = string_utils.quote(max_value)

    sql_text = sql_text.format(column=quoted_name(column, True),
                               min=min_value, max=max_value)

    options.update(sqltext=sql_text)
    return CheckConstraint(**options)


def get_class_by_table(base, table, **options):
    """
    gets declarative class associated with given table.

    if no class is found this function returns `None`.
    if multiple classes were found (polymorphic cases) additional `data` parameter
    can be given to hint which class to return.

    :param type[BaseEntity] base: declarative base model.
    :param Table table: sqlalchemy table object.

    :keyword dict data: data row to determine the class in polymorphic scenarios.

    :keyword bool raise_multi: specifies that if multiple classes found and
                               also provided data could not help, raise an error.
                               otherwise return None.
                               defaults to True if not provided.

    :note: this code is taken from sqlalchemy-utils project.
    https://github.com/kvesteri/sqlalchemy-utils

    for example:

    class User(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()
        name = StringColumn()

    get_class_by_table(CoreEntity, User.__table__) -> User class

    this function also supports models using single table inheritance.
    additional data parameter should be provided in these cases.

    for example:

    class Entity(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()
        name = StringColumn()
        type = StringColumn()
        __mapper_args__ = {
            'polymorphic_on': type,
            'polymorphic_identity': 'entity'
        }

    class User(Entity):
        __mapper_args__ = {
            'polymorphic_identity': 'user'
        }

    get_class_by_table(CoreEntity, Entity.__table__, {'type': 'entity'}) -> Entity class
    get_class_by_table(CoreEntity, Entity.__table__, {'type': 'user'}) -> User class

    it also supports extended entities with unlimited depth, it returns the correct child
    entity.

    for example:

    class EntityBase(CoreEntity):
        _table = 'entity'
        id = AutoPKColumn()

    class Entity2(EntityBase):
        _extend_existing = True
        name = StringColumn()

    class Entity3(Entity2):
        _extend_existing = True
        age = CoreColumn(Integer)

    get_class_by_table(CoreEntity, EntityBase.__table__) -> Entity3 class

    :raises MultipleDeclarativeClassesFoundError: multiple declarative classes found error.

    :returns: declarative class or None.
    :rtype: type[pyrin.database.model.base.BaseEntity]
    """

    data = options.get('data')
    raise_multi = options.get('raise_multi', True)
    found_classes = []
    for item in base.registry.mappers:
        if len(item.tables) > 0 and item.tables[0] is table:
            found_classes.append(item.entity)

    current_count = len(found_classes)
    temp = list(found_classes)
    if current_count > 1:
        for item in found_classes:
            if item.extend_existing is not True:
                temp.remove(item)

    if current_count > len(temp) > 0:
        found_classes = temp
        if len(found_classes) > 1:
            hierarchies = []
            for item in found_classes:
                hierarchies.append(set(item.__mro__))

            hierarchies = sorted(hierarchies, key=len, reverse=True)
            last = hierarchies[0]
            others = hierarchies[1:]
            union = set().union(*others)
            result = union.symmetric_difference(last)
            if len(result) == 1:
                return result.pop()

        elif len(found_classes) == 1:
            return found_classes[0]

    found_classes = set(found_classes)
    if len(found_classes) > 1:
        if not data:
            if raise_multi is True:
                raise MultipleDeclarativeClassesFoundError('Multiple declarative classes found '
                                                           'for table [{table}]. please provide '
                                                           '"data" parameter for this function '
                                                           'to be able to determine polymorphic '
                                                           'scenarios.'.format(table=table.name))
        else:
            for cls in found_classes:
                mapper = inspect(cls)
                polymorphic_on = mapper.polymorphic_on.name
                if polymorphic_on in data:
                    if data[polymorphic_on] == mapper.polymorphic_identity:
                        return cls

            if raise_multi is True:
                raise MultipleDeclarativeClassesFoundError('Multiple declarative classes found '
                                                           'for table [{table}]. given data row '
                                                           'does not match any polymorphic '
                                                           'identity of the found classes.'
                                                           .format(table=table.name))
    elif found_classes:
        return found_classes.pop()

    return None


def is_valid_column_name(column):
    """
    gets a value indicating that given column name is valid.

    :param str column: column name.

    :rtype: bool
    """

    if not isinstance(column, str) or len(column) <= 0:
        return False

    column = column.replace('+', '').replace('-', '').strip()
    return len(column) > 0 and ' ' not in column


def get_column_name(column):
    """
    gets the pure column name from given column name with ordering info.

    for example:

    +age -> age
    age -> age
    -age -> age

    :param str column: column name to extract pure name from it.

    :rtype: str
    """

    if column.startswith(('-', '+')):
        return column[1:]

    return column


def get_ordering_info(column):
    """
    gets a tuple containing ordering info for given column name.

    it returns a tuple of two item, first item is column name and
    second item is ordering type which is an `UnaryExpression` from `asc` or `desc`.

    default ordering is ascending, but it could be changed to descending
    by prefixing `-` to column name.

    for example:

    age -> ordering for age column ascending.
    -age -> ordering for age column descending.

    :param str column: column name to get its ordering info.

    :rtype: tuple[str, UnaryExpression]
    """

    order_type = asc
    if column.startswith(('-', '+')):
        if column[0] == '-':
            order_type = desc
        column = column[1:]

    return column, order_type


def get_ordering_criterion(*columns, valid_columns=None, ignore_invalid=True):
    """
    gets required criterion for given columns ordering.

    default ordering is ascending, but it could be changed to descending
    by prefixing `-` to column names.

    this method always ignores empty strings and None values.

    for example:

    name, +age -> ordering for name and age columns both ascending.
    name, -age -> ordering for name ascending and age descending.

    :param str columns: column names to get their ordering criterion.

    :param list[str] valid_columns: valid column names for ordering.
                                    defaults to None if not provided.

    :param bool ignore_invalid: specifies that if provided columns are
                                not in valid column names, ignore them
                                instead of raising an error. note that
                                this only has effect if `valid_columns`
                                is provided. defaults to True.

    :raises InvalidOrderingColumnError: invalid ordering column error.

    :rtype: tuple[UnaryExpression]
    """

    result = []
    error_message = _('Column [{name}] is not valid for ordering.')
    for item in columns:
        if is_valid_column_name(item):
            name, order_type = get_ordering_info(item)
            if valid_columns is None or name in valid_columns:
                result.append(order_type(quoted_name(name, True)))
            elif ignore_invalid is False:
                raise InvalidOrderingColumnError(error_message.format(name=name))

        elif ignore_invalid is False:
            raise InvalidOrderingColumnError(error_message.format(name=item))

    return tuple(result)


def is_expression_level_hybrid_property(value):
    """
    gets a value indicating that provided object is an expression level hybrid property.

    the provided value may be a proxy holding a descriptor to a hybrid property.

    :param object value: value to be checked.

    :rtype: bool
    """

    if isinstance(value, hybrid_property):
        return value.expr is not None

    descriptor = getattr(value, 'descriptor', None)
    if descriptor is None:
        return False

    if not isinstance(descriptor, hybrid_property):
        return False

    return descriptor.expr is not None



# -*- coding: utf-8 -*-
"""
authorization handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class AuthorizationHandlersException(CoreException):
    """
    authorization handlers exception.
    """
    pass


class AuthorizationHandlersBusinessException(CoreBusinessException,
                                             AuthorizationHandlersException):
    """
    authorization handlers business exception.
    """
    pass


class AuthorizerNameIsRequiredError(AuthorizationHandlersException):
    """
    authorizer name is required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
admin page exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.validator.exceptions import ValidationError


class AdminPageException(CoreException):
    """
    admin page exception.
    """
    pass


class AdminPageBusinessException(CoreBusinessException,
                                 AdminPageException):
    """
    admin page business exception.
    """
    pass


class InvalidListFieldError(AdminPageException):
    """
    invalid list field error.
    """
    pass


class InvalidListSearchFieldError(AdminPageException):
    """
    invalid list search field error.
    """
    pass


class ListFieldRequiredError(AdminPageException):
    """
    list field required error.
    """
    pass


class InvalidMethodNameError(AdminPageException):
    """
    invalid method name error.
    """
    pass


class InvalidAdminEntityTypeError(AdminPageException):
    """
    invalid admin entity type error.
    """
    pass


class AdminRegisterNameRequiredError(AdminPageException):
    """
    admin register name required error.
    """
    pass


class AdminNameRequiredError(AdminPageException):
    """
    admin name required error.
    """
    pass


class RequiredValuesNotProvidedError(ValidationError, AdminPageBusinessException):
    """
    required values not provided error.
    """
    pass


class CompositePrimaryKeysNotSupportedError(AdminPageException):
    """
    composite primary keys not supported error.
    """
    pass


class DuplicateListFieldNamesError(AdminPageException):
    """
    duplicate list field names error.
    """
    pass


class DuplicateListSearchFieldNamesError(AdminPageException):
    """
    duplicate list search field names error.
    """
    pass


class InvalidListFieldNameError(AdminPageException):
    """
    invalid list field name error.
    """
    pass


class EntityNotFoundError(AdminPageBusinessException):
    """
    entity not found error.
    """
    pass


class ExtraDataFieldsAndEntityFieldsOverlapError(AdminPageException):
    """
    extra data fields and entity fields overlap error.
    """
    pass



# -*- coding: utf-8 -*-
"""
admin page mixin module.
"""

from pyrin.caching.mixin.typed import TypedCacheMixin


class AdminPageCacheMixin(TypedCacheMixin):
    """
    admin page cache mixin class.

    this class adds caching support to its subclasses.
    """

    _container = {}



# -*- coding: utf-8 -*-
"""
swagger interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class TagSingletonMeta(MultiSingletonMeta):
    """
    tag singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractTag(CoreObject, metaclass=TagSingletonMeta):
    """
    abstract tag class.
    """

    # name of this tag to be registered with.
    _name = None

    # tag name of this tag to be used in swagger ui.
    _tag = None

    @abstractmethod
    def is_accepted(self, rule, method, **options):
        """
        gets a value indicating that this rule is accepted for this tag.

        :param pyrin.api.router.handlers.base.RouteBase rule: rule instance to be processed.
        :param str method: http method name.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        gets the name of this tag.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        return self._name

    @property
    @abstractmethod
    def tag(self):
        """
        gets the tag name of this tag.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        return self._tag



# -*- coding: utf-8 -*-
"""
session enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class RequestContextEnum(CoreEnum):
    """
    request context enum.
    """

    PAGINATOR = 'paginator'
    RESULT_SCHEMA = 'result_schema'



# -*- coding: utf-8 -*-
"""
authorization decorators module.
"""

import pyrin.security.authorization.services as authorization_services


def authorizer(*args, **kwargs):
    """
    decorator to register an authorizer.

    :param object args: authorizer class constructor arguments.
    :param object kwargs: authorizer class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           authorizer with the same name, replace it
                           with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidAuthorizerTypeError: invalid authorizer type error.
    :raises DuplicatedAuthorizerError: duplicated authorizer error.

    :returns: authorizer class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available authorizer.

        :param type cls: authorizer class.

        :returns: authorizer class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        authorization_services.register_authorizer(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
audit api module.
"""

import pyrin.audit.services as audit_services

from pyrin.api.router.decorators import api


audit_config = audit_services.get_audit_configurations()
audit_config.update(no_cache=True)
is_enabled = audit_config.pop('enabled', False)

if is_enabled is True:
    @api(**audit_config)
    def inspect(**options):
        """
        inspects all registered packages and gets inspection data.
        ---
        parameters:
          - name: application
            type: boolean
            description: specifies that application info must be included
          - name: packages
            type: boolean
            description: specifies that loaded packages info must be included
          - name: framework
            type: boolean
            description: specifies that framework info must be included
          - name: python
            type: boolean
            description: specifies that python info must be included
          - name: os
            type: boolean
            description: specifies that operating system info must be included
          - name: hardware
            type: boolean
            description: specifies that hardware info must be included
          - name: database
            type: boolean
            description: specifies that database info must be included
          - name: caching
            type: boolean
            description: specifies that caching info must be included
          - name: celery
            type: boolean
            description: specifies that celery info must be included
          - name: traceback
            type: boolean
            description: specifies that on failure, it must include the traceback of errors
        responses:
          200:
            description: all packages are working normally
            schema:
              properties:
                application:
                  type: object
                  description: application info
                packages:
                  type: object
                  description: loaded packages info
                framework:
                  type: object
                  description: framework info
                python:
                  type: object
                  description: python info
                platform:
                  type: object
                  description: platform info
                database:
                  type: object
                  description: database info
                caching:
                  type: object
                  description: caching info
                celery:
                  type: object
                  description: celery info
          500:
            description: some packages have errors
            schema:
              properties:
                application:
                  type: object
                  description: application info
                packages:
                  type: object
                  description: loaded packages info
                framework:
                  type: object
                  description: framework info
                python:
                  type: object
                  description: python info
                platform:
                  type: object
                  description: platform info
                database:
                  type: object
                  description: database info
                caching:
                  type: object
                  description: caching info
                celery:
                  type: object
                  description: celery info
        """

        return audit_services.inspect(**options)



# -*- coding: utf-8 -*-
"""
users internal models module.
"""

from pyrin.database.model.mixin import HistoryMixin
from pyrin.database.model.declarative import CoreEntity
from pyrin.users.mixin import UserPKMixin, UserCommonMixin, UserAccessMixin, \
    UserCredentialMixin


class InternalUserEntity(CoreEntity, UserPKMixin,
                         UserCredentialMixin, UserCommonMixin,
                         UserAccessMixin, HistoryMixin):
    """
    internal user entity class.
    """

    _table = 'internal_user'



# -*- coding: utf-8 -*-
"""
session manager module.
"""

from flask import request
from flask.ctx import has_request_context

import pyrin.security.authentication.services as authentication_services
import pyrin.security.authorization.services as authorization_services

from pyrin.core.structs import Manager
from pyrin.security.session import SessionPackage
from pyrin.security.session.exceptions import InvalidUserError, \
    CouldNotOverwriteCurrentUserError, InvalidComponentCustomKeyError


class SessionManager(Manager):
    """
    session manager class.
    """

    package_class = SessionPackage

    def get_current_user(self):
        """
        gets current user identity.

        :returns: object
        """

        return self.get_current_request().user

    def get_current_user_info(self):
        """
        gets current user info.

        it may return None if user info is not set.

        :rtype: dict
        """

        return self.get_current_request().user_info

    def set_current_user(self, user, info=None):
        """
        sets current user.

        :param user: user identity object.
        :param dict info: user info.

        :raises InvalidUserError: invalid user error.
        :raises CouldNotOverwriteCurrentUserError: could not overwrite current user error.
        """

        if user is None:
            raise InvalidUserError('Request user could not be None.')

        current_request = self.get_current_request()
        if current_request.user is not None:
            raise CouldNotOverwriteCurrentUserError('User has been already set for '
                                                    'current request, it could not '
                                                    'be overwritten.')

        current_request.user = user
        current_request.user_info = info

    def get_current_request(self):
        """
        gets current request object.

        :raises RuntimeError: runtime error.

        :rtype: pyrin.processor.request.wrappers.base.CoreRequest
        """

        return request

    def get_current_request_id(self):
        """
        gets current request id.

        this method is implemented to overcome the hash problem of requests.
        `CoreRequest` objects are hashable themselves, but when they used as a dict key
        some hash collisions will occur. so we have to expose the exact request id to
        be able to use it as a dict key in places such as database scoped sessions.

        :raises RuntimeError: runtime error.

        :rtype: uuid.UUID
        """

        return self.get_current_request().request_id

    def add_request_context(self, key, value, **options):
        """
        adds the given key/value pair into current request context.

        :param str key: key name to be added.
        :param object value: value to be added.

        :keyword bool replace: specifies that if a key with the same name
                               is already present, replace it. otherwise
                               raise an error. defaults to False if not provided.

        :raises InvalidRequestContextKeyNameError: invalid request context key name error.
        :raises RequestContextKeyIsAlreadyPresentError: request context key is
                                                        already present error.
        """

        self.get_current_request().add_context(key, value, **options)

    def get_request_context(self, key, default=None):
        """
        gets the value for given key from current request context.

        it gets the default value if key is not present in the request context.

        :param str key: key name to get its value.
        :param object default: a value to be returned if the provided
                               key is not present in request context.

        :returns: object
        """

        return self.get_current_request().get_context(key, default)

    def remove_request_context(self, key):
        """
        removes the specified key from current request context if available.

        :param str key: key name to be removed from request context.
        """

        self.get_current_request().remove_context(key)

    def is_fresh(self):
        """
        gets a value indicating that current request has a fresh authentication.

        fresh authentication means an authentication which is done by providing
        user credentials to server.

        :rtype: bool
        """

        authenticator = authentication_services.get_current_authenticator()
        if authenticator is not None:
            return authenticator.is_fresh()

        return False

    def set_component_custom_key(self, value):
        """
        sets the component custom key.

        :param object value: component custom key value.

        :raises InvalidComponentCustomKeyError: invalid component custom key error.
        """

        if value is None:
            raise InvalidComponentCustomKeyError('Component custom key could not be None.')

        self.get_current_request().component_custom_key = value

    def get_component_custom_key(self):
        """
        gets component custom key.

        :rtype: object
        """

        return self.get_current_request().component_custom_key

    def get_safe_component_custom_key(self):
        """
        gets component custom key in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.

        :rtype: object
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.component_custom_key

    def get_safe_current_request(self):
        """
        gets current request object in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.

        :rtype: pyrin.processor.request.wrappers.base.CoreRequest
        """

        if self.is_request_context_available() is True:
            return self.get_current_request()

        return None

    def get_safe_current_user(self):
        """
        gets current user in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.user

    def get_safe_cacheable_current_user(self):
        """
        gets cacheable current user in a safe manner.

        meaning that if the request does not exist in current context, it will
        return a None object instead of raising an error.
        """

        current_request = self.get_safe_current_request()
        if current_request is None:
            return None

        return current_request.cacheable_user

    def is_request_context_available(self):
        """
        gets a value indicating that request context is available for usage.

        :rtype: bool
        """

        return has_request_context()

    def is_superuser(self):
        """
        gets a value indicating that the current user is superuser.

        :rtype: bool
        """

        authorizer_name = authentication_services.get_current_authenticator_name()
        if authorizer_name is None or \
                not authorization_services.authorizer_exists(authorizer_name):
            return False

        authorizer = authorization_services.get_authorizer(authorizer_name)
        return authorizer.is_superuser()

    def set_response_cookie(self, key, value, path='/',
                            secure=False, httponly=False, **options):
        """
        sets a response cookie that client should send on subsequent requests.

        :param str key: the key (name) of the cookie to be set.
        :param str value: the value of the cookie.

        :param str path: limits the cookie to a given path, per default
                         it will span the whole domain.

        :param bool secure: if `True`, the cookie will only be available via HTTPS.
        :param bool httponly: disallow JavaScript access to the cookie.

        :keyword timedelta | int max_age: should be a number of seconds, or `None`
                                          (default) if the cookie should last only
                                          as long as the client's browser session.

        :keyword str | datetime | int | float expires: should be a `datetime`
                                                       object or UNIX timestamp.

        :keyword str domain: if you want to set a cross-domain cookie.
                             for example, `domain=".example.com"` will set a
                             cookie that is readable by the domain `www.example.com`,
                             `foo.example.com`` etc. otherwise, a cookie will only
                             be readable by the domain that set it.

        :keyword samesite: limit the scope of the cookie to only be
                           attached to requests that are `same-site`.

        :enum samesite:
            STRICT = 'Strict'
            LAX = 'Lax'
            NONE = 'None'
        """

        current_request = self.get_current_request()
        current_request.set_response_cookie(key, value, path,
                                            secure, httponly, **options)



# -*- coding: utf-8 -*-
"""
swagger enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class ParameterLocationEnum(CoreEnum):
    """
    parameter location enum.
    """

    PATH = 'path'
    QUERY = 'query'
    BODY = 'body'
    HEADER = 'header'
    FORM = 'formData'


class ParameterAttributeEnum(CoreEnum):
    """
    parameter attribute enum.
    """

    # name of the parameter.
    NAME = 'name'

    # description of the parameter.
    DESCRIPTION = 'description'

    # specifies that parameter is required.
    REQUIRED = 'required'

    # type of the parameter.
    # it could be from 'ParameterTypeEnum' values.
    TYPE = 'type'

    # the place of this parameter in request.
    # it could be from 'ParameterLocationEnum' values.
    IN = 'in'

    # the format of parameter for given type.
    # it could be from 'ParameterFormatEnum' values.
    FORMAT = 'format'

    # list of valid values for given parameter.
    ENUM = 'enum'

    # items of an array type parameter.
    ITEMS = 'items'

    # keys of an object type parameter.
    PROPERTIES = 'properties'

    # default value for parameter.
    DEFAULT = 'default'

    # minimum value for parameter.
    MINIMUM = 'minimum'

    # maximum value for parameter.
    MAXIMUM = 'maximum'

    # schema of parameter.
    SCHEMA = 'schema'


class ParameterFormatEnum(CoreEnum):
    """
    parameter format enum.
    """

    UUID = 'uuid'
    EMAIL = 'email'
    DATE = 'date'
    TIME = 'time'
    DATE_TIME = 'date-time'
    PASSWORD = 'password'
    BYTE = 'byte'
    URI = 'uri'
    HOSTNAME = 'hostname'
    IPV4 = 'ipv4'
    IPV6 = 'ipv6'
    DOUBLE = 'double'
    FLOAT = 'float'
    TEXT = 'text'


class ParameterTypeEnum(CoreEnum):
    """
    parameter type enum.
    """

    INTEGER = 'integer'
    NUMBER = 'number'
    BOOLEAN = 'boolean'
    STRING = 'string'
    ARRAY = 'array'
    OBJECT = 'object'


class DocstringSectionEnum(CoreEnum):
    """
    docstring section enum.
    """

    # list of parameters.
    PARAMETERS = 'parameters'

    # dict of responses.
    RESPONSES = 'responses'

    # list of security definitions.
    SECURITY = 'security'

    # list of tags for each api to be categorized by tag in swagger ui.
    TAGS = 'tags'
